from unittest import TestCase

import numpy as np

import remotior_sensus
from remotior_sensus.util import files_directories


class TestBandSetCatalog(TestCase):

    def test_bandset_catalog(self):
        rs = remotior_sensus.Session(
            n_processes=2, available_ram=1000, log_level=10
            )
        cfg = rs.configurations
        cfg.logger.log.debug('test')
        data_directory = './data/S2_2020-01-01'
        cfg.logger.log.debug('>>> test bandset catalog create')
        # create BandSet Catalog
        catalog = rs.bandset_catalog()
        # BandSet count
        self.assertEqual(catalog.get_bandset_count(), 1)
        # empty BandSet
        catalog.create_bandset(insert=True, bandset_name='BandSet X')
        # set box_coordinate_list
        catalog.get(1).box_coordinate_list = [1, 1, 1, 1]
        # BandSet count
        self.assertEqual(catalog.get_bandset_count(), 2)
        self.assertEqual(
            catalog.get(1).box_coordinate_list[0],
            catalog.bandsets_table['box_coordinate_left'][
                catalog.bandsets_table.bandset_number == 1]
        )
        # set name
        new_name = 'new_name'
        catalog.get(1).name = new_name
        self.assertEqual(
            new_name, catalog.bandsets_table['bandset_name'][
                catalog.bandsets_table.bandset_number == 1]
        )
        # set crs
        new_crs = 'crs'
        catalog.get(1).crs = new_crs
        self.assertEqual(
            new_crs, catalog.bandsets_table['crs'][
                catalog.bandsets_table.bandset_number == 1]
        )
        # set date
        catalog.get(1).date = '2022-01-01'
        self.assertEqual(
            catalog.get(1).date, catalog.bandsets_table['date'][
                catalog.bandsets_table.bandset_number == 1]
        )
        # set root
        catalog.get(1).root_directory = '/dir'
        self.assertEqual(
            catalog.get(1).root_directory,
            catalog.bandsets_table['root_directory'][
                catalog.bandsets_table.bandset_number == 1]
        )
        # clear BandSet
        catalog.clear_bandset(bandset_number=1)
        # BandSet from directory
        catalog.create_bandset(
            paths=[data_directory, 'tif'], wavelengths=['Sentinel-2'],
            date=cfg.date_auto, bandset_number=1, bandset_name='bandset_1'
        )
        self.assertEqual(catalog.get_bandset_count(), 2)

        catalog.set_satellite_wavelength(
            satellite_name=cfg.no_satellite, bandset_number=1
        )
        self.assertEqual(catalog.get(1).bands[0].wavelength, 1)
        wavelength_list = [
            0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2
        ]
        catalog.set_wavelength(
            wavelength_list=wavelength_list, unit=cfg.wl_micro,
            bandset_number=1
        )
        self.assertEqual(catalog.get(1).bands[0].wavelength, 0.1)
        catalog.set_satellite_wavelength(
            satellite_name=cfg.satSentinel2, bandset_number=1
        )
        self.assertEqual(
            catalog.get(1).bands[0].wavelength,
            cfg.satellites[cfg.satSentinel2][0][0]
        )
        file_list = ['S2_2020-01-01/S2_B02.tif', 'S2_2020-01-01/S2_B03.tif',
                     'S2_2020-01-01/S2_B04.tif']
        date = '2021-01-01'
        root_directory = './data'
        catalog.create_bandset(
            file_list, wavelengths=['Sentinel-2'], date=date, bandset_number=2,
            root_directory=root_directory
        )
        # find BandSet from name list
        name_list = ['ba', 'S2_B']
        bandset_list = catalog.find_bandset_names_in_list(names=name_list)
        self.assertEqual(len(bandset_list), 2)
        bandset_list = catalog.find_bandset_names_in_list(
            names=name_list, exact_match=True
        )
        self.assertEqual(len(bandset_list), 1)
        b_paths = catalog.get_band_list(bandset=2, bandset_catalog=catalog)
        d_paths = ['1', '2', '3']
        catalog.set_paths(path_list=d_paths, bandset_number=2)
        self.assertNotEqual(
            catalog.get_band_list(bandset=2, bandset_catalog=catalog), b_paths
        )
        catalog.set_paths(path_list=file_list, bandset_number=2)
        self.assertEqual(
            catalog.get_band_list(bandset=2, bandset_catalog=catalog), b_paths
        )
        # box coordinate list
        coordinate_list = [230250, 4674550, 230320, 4674440]
        catalog.create_bandset(
            file_list, wavelengths=['Sentinel-2'], date=date, bandset_number=3,
            root_directory=root_directory, box_coordinate_list=coordinate_list
        )
        self.assertEqual(
            catalog.get(3).box_coordinate_list,
            catalog.get_box_coordinate_list(3)
        )
        self.assertEqual(coordinate_list, catalog.get_box_coordinate_list(3))
        # export as xml
        temp = cfg.temp.temporary_file_path(name_suffix='.xml')
        catalog.export_bandset_as_xml(2, output_path=temp)
        self.assertTrue(files_directories.is_file(temp))
        catalog.import_bandset_from_xml(3, xml_path=temp)
        for i in range(len(catalog.get(2).bands[0])):
            if str(catalog.get(2).bands[0][i]) != 'NaT':
                self.assertEqual(catalog.get(2).bands[0][i],
                                 catalog.get(3).bands[0][i])
        self.assertEqual(
            [catalog.get(2).date, catalog.get(2).root_directory,
             catalog.get(2).crs, catalog.get(2).name,
             catalog.get(2).box_coordinate_list],
            [catalog.get(3).date, catalog.get(3).root_directory,
             catalog.get(3).crs, catalog.get(3).name,
             catalog.get(3).box_coordinate_list])
        # set current BandSet
        catalog.current_bandset = 10
        # current BandSet
        self.assertEqual(catalog.current_bandset, catalog.get_bandset_count())
        # create second catalog
        catalog_2 = rs.bandset_catalog()
        self.assertEqual(str(catalog_2.bandsets_table['date'][0]), 'NaT')
        # BandSet count
        self.assertEqual(catalog.get_bandset_count(), 3)
        self.assertEqual(catalog_2.get_bandset_count(), 1)
        # get BandSet function
        bs = catalog.get_bandset
        # get BandSet
        self.assertEqual(bs(1), catalog.get_bandset(1))
        self.assertEqual(type(catalog.get_bandset(2)), rs.bandset)
        self.assertGreater(len(str(catalog.get_bandset(1).crs)), 0)
        # get bands
        self.assertEqual(
            bs(2).bands['name'].tolist(),
            catalog.get_bandset(2).bands['name'].tolist()
        )
        self.assertEqual(type(catalog.get_bandset(1).bands), np.recarray)
        # BandSet absolute paths
        self.assertEqual(
            bs(2).get_absolute_paths()[0],
            catalog.get_bandset(2).get_absolute_path(band_number=1)
        )
        # BandSet relative paths
        self.assertEqual(bs(2).get_paths(), bs(2, 'path'))
        self.assertEqual(
            catalog.get_bandset(2).get_paths(), catalog.get_bandset(2, 'path')
        )
        self.assertEqual(
            catalog.get_bandset(2).get_band_attributes('path'), file_list
        )
        # BandSet date
        self.assertEqual(str(bs(2, 'date')[0]), date)
        self.assertEqual(bs(2).date, catalog.get_bandset(2).date)
        self.assertEqual(
            bs(2).date, catalog.get_bandset_catalog_attributes(2, 'date')
        )
        self.assertEqual(str(catalog.get_bandset(2, 'date')[0]), date)
        date_current = catalog.get_bandset_catalog_attributes(
            catalog.current_bandset, 'date'
        )
        self.assertEqual(
            date_current, catalog.get_bandset(
                catalog.current_bandset
            ).get_band_attributes('date')[0]
        )
        # bandset calculation
        calc = bs(3).calc('"b1" + "b2"')
        self.assertTrue(files_directories.is_file(calc.paths[0]))
        # bandset calculation
        catalog.get_bandset(3).calc(
            '"b1" / "b2"', point_coordinates=[230274, 4674515]
        )
        cfg.multiprocess.multiprocess_pixel_value()
        value = cfg.multiprocess.output
        self.assertTrue(value > 0)
        # BandSet wavelengths
        self.assertEqual(
            bs(2).get_wavelengths(), catalog.get_bandset(2).get_wavelengths()
        )
        names = catalog.iterate_bandset_bands('name')
        self.assertGreater(len(names), 0)
        # iterate BandSets
        b_list = catalog.get_bandsets_by_list()
        self.assertEqual(len(b_list), catalog.get_bandset_count())
        b_list = catalog.get_bandsets_by_list(output_number=True)
        self.assertEqual(len(b_list), catalog.get_bandset_count())
        b_list = catalog.get_bandsets_by_list(bandset_list=[1, 3])
        self.assertEqual(len(b_list), 2)
        # get bandsets by date
        bandsets = catalog.get_bandsets_by_date(
            date_list=['>=2020-01-01 & <=2022-01-02', '2020-01-01'],
            output_number=True
        )
        self.assertEqual(len(bandsets), 3)
        # move band in BandSet
        band_3 = catalog.get_bandset(1).get_band(3).name
        catalog.move_band_in_bandset(
            bandset_number=1, band_number_input=3, band_number_output=2
        )
        self.assertEqual(band_3, catalog.get_bandset(1).get_band(2).name[0])
        band_4 = catalog.get_bandset(1).get_band(4).name
        catalog.move_band_in_bandset(
            bandset_number=1, band_number_input=4, band_number_output=6
        )
        self.assertEqual(band_4, catalog.get_bandset(1).get_band(6).name[0])
        # add band to BandSet
        bandset_1_count = catalog.get_bandset(1).get_band_count()
        catalog.add_band_to_bandset(
            path='./data/S2_2020-01-01/S2_B02.tif', bandset_number=1,
            band_number=1, raster_band=1
        )
        self.assertGreater(
            catalog.get_bandset(1).get_band_count(), bandset_1_count
        )
        # test band overview creation
        band_list = catalog.get_bandset(1).get_absolute_paths()
        temp_file = cfg.temp.temporary_file_path(name_suffix='.tif')
        files_directories.copy_file(in_path=band_list[0], out_path=temp_file)
        # load BandSet catalog
        catalog_3 = rs.bandset_catalog()
        catalog_3.create_bandset(paths=[temp_file], bandset_number=1)
        catalog_3.build_bandset_band_overview(1)
        self.assertTrue(files_directories.is_file(temp_file + '.ovr'))
        band_1_name = catalog.get(1).get(1).name
        # sort bands by wavelength
        catalog.sort_bands_by_wavelength(bandset_number=1)
        # get band by nearest wavelength
        band_1 = catalog.get_bandset(1).get_band_by_wavelength(wavelength=1)
        self.assertEqual(band_1.name, band_1_name)
        # get BandSet band from attribute
        band_w = catalog.get_bandset_bands_by_attribute(
            bandset_number=1, attribute='wavelength',
            attribute_value=catalog.get(1).get(1).wavelength
        )
        self.assertEqual(
            band_w[0].wavelength, catalog.get_bandset(1).get_band(1).wavelength
        )
        # remove band in BandSet
        band_number = catalog.get_bandset(1).get_band_by_wavelength(
            1, output_as_number=True
        )
        catalog.remove_band_in_bandset(
            bandset_number=1, band_number=band_number
        )
        self.assertEqual(
            catalog.get_bandset(1).get_band_count(), bandset_1_count
        )
        # get BandSet by number
        bandset_by_number = catalog.get_bandset_by_number(1)
        self.assertEqual(bandset_by_number.get_band_count(), bandset_1_count)
        # get BandSet by name
        bandset_by_name = catalog.get_bandset_by_name('bandset_1')
        self.assertEqual(bandset_by_name.get_band_count(), bandset_1_count)
        # get band list
        band_list = rs.bandset_catalog.get_band_list(1, catalog)
        self.assertGreater(len(band_list), 1)
        # load BandSet catalog
        empty_catalog = rs.bandset_catalog()
        self.assertEqual(empty_catalog.get_bandset_count(), 1)
        empty_catalog._load(bandset_catalog=catalog, current_bandset=1)
        self.assertEqual(
            empty_catalog.get_bandset_count(), catalog.get_bandset_count()
        )
        empty_catalog.add_bandset(
            catalog.get_bandset(1), bandset_number=1, insert=True
        )
        self.assertEqual(
            empty_catalog.get_bandset_count(), catalog.get_bandset_count() + 1
        )
        empty_catalog.move_bandset(
            bandset_number_input=1, bandset_number_output=3
        )
        self.assertEqual(
            empty_catalog.get_bandset(3).get_band_attributes('name'),
            catalog.get_bandset(1).get_band_attributes('name')
        )
        # set BandSet root directory
        empty_catalog.set_root_directory(
            bandset_number=1, root_directory=data_directory
        )
        self.assertEqual(
            empty_catalog.get(1).root_directory,
            empty_catalog.get_root_directory(1)
        )
        # set BandSet date
        empty_catalog.set_date(bandset_number=1, date=date)
        self.assertEqual(
            str(empty_catalog.get(1).date),
            empty_catalog.get_date(1)
            )
        # BandSet band count
        self.assertEqual(
            empty_catalog.get_band_count(bandset_number=1),
            catalog.get_band_count(bandset_number=1)
        )
        # set BandSet name
        empty_catalog.set_name(bandset_number=1, name='new_name')
        self.assertEqual(empty_catalog.get(1).name, empty_catalog.get_name(1))
        # set BandSet box coordinate list
        empty_catalog.set_box_coordinate_list(
            bandset_number=1, box_coordinate_list=coordinate_list
        )
        self.assertEqual(
            empty_catalog.get_box_coordinate_list(1), coordinate_list
        )
        band_string = empty_catalog.create_band_string_list(bandset_number=1)
        self.assertGreater(len(band_string), 0)
        # remove BandSets
        empty_catalog.remove_bandset(1)
        empty_catalog.remove_bandset(1)
        empty_catalog.remove_bandset(1)
        empty_catalog.remove_bandset(1)
        self.assertEqual(empty_catalog.get_bandset(1).get_band_count(), 0)
        # set date
        catalog.get(2).date = '2022-01-01'
        catalog.sort_bandsets_by_date()
        self.assertEqual(catalog.get_date(3), '2022-01-01')
        catalog.sort_bands_by_name(bandset_number=1)
        # print bandset 1
        catalog.print_bandset(1)
        # update crs
        catalog.get(1).crs = None
        catalog.update_crs(bandset_number=1)
        self.assertTrue(catalog.get(1).crs is not None)
        histogram = catalog.calculate_scatter_plot_histogram(
            bandset_number=1, band_x=1, band_y=2,
            vector_path='./data/files/roi.gpkg'
        )
        self.assertTrue(histogram is not None)
        # create virtual raster
        virtual = catalog.create_virtual_raster(bandset_number=1)
        self.assertTrue(files_directories.is_file(virtual))
        stack = catalog.create_bandset_stack(bandset_number=1)
        self.assertTrue(files_directories.is_file(stack))

        # clear temporary directory
        rs.close()



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""Band classification.

This tool allows for the classification of remote sensing images, providing
several algorithms such as Minimum Distance, Maximum Likelihood,
Spectral Angle Mapping.
Also, machine learning algorithms are provided through
`PyTorch <https://pytorch.org/>`_ (pytorch_multi_layer_perceptron) and
`scikit-learn <https://scikit-learn.org/stable/>`_ (random_forest,
random_forest_ovr, support_vector_machine, multi_layer_perceptron).
This module includes tools for training the algorithms using
Regions of Interest (ROIs) or spectral signatures.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # start the process
    >>> classification = rs.band_classification(
    ... input_bands=['file1.tif', 'file2.tif'], output_path='output.tif',
    ... algorithm_name=cfg.maximum_likelihood
    ... )
"""

import pickle
from copy import deepcopy
from typing import Union, Optional

import numpy as np

from remotior_sensus.core import configurations as cfg
from remotior_sensus.core.bandset_catalog import BandSet
from remotior_sensus.core.bandset_catalog import BandSetCatalog
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.core.processor_functions import (
    classification_maximum_likelihood, classification_minimum_distance,
    classification_spectral_angle_mapping, classification_scikit,
    classification_pytorch, get_band_arrays, fit_classifier,
    score_classifier_stratified, score_classifier
)
from remotior_sensus.core.spectral_signatures import SpectralSignaturesCatalog
from remotior_sensus.util import (
    files_directories, raster_vector, shared_tools, read_write_files
)

try:
    import torch
    from remotior_sensus.util.pytorch_tools import train_pytorch_model
except Exception as error:
    if cfg.logger is not None:
        cfg.logger.log.error(str(error))

try:
    from sklearn import svm
    from sklearn.base import clone
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.model_selection import StratifiedKFold
    from sklearn.multiclass import OneVsRestClassifier
    from sklearn.neural_network import MLPClassifier
except Exception as error:
    if cfg.logger is not None:
        cfg.logger.log.error(str(error))
    else:
        print(str(error))


class Classifier(object):
    """Manages classifiers.

    A classifier is an object which includes the required parameters to perform 
    a classification, including the tools to perform the training.

    Attributes:
        algorithm_name: algorithm name selected form cfg.classification_algorithms.
        spectral_signatures: a SpectralSignaturesCatalog containing spectral signatures.
        covariance_matrices: dictionary of previously calculated covariance matrices 
            (used in maximum_likelihood).
        model_classifier: classifier object.
        input_normalization: perform input normalization; options are z_score or linear_scaling.
        normalization_values: list of normalization paramters defined for each variable
            [normalization expressions, mean values, 
            standar deviation values, minimum values, maximum values].
        framework_name: name of framework such as 
            classification_framework, scikit_framework, or pytorch_framework.
        classification_function: the actual classification function.
        function_argument = a dictionary including arguments for the classification function
            such as model_classifier, covariance_matrices, normalization_values, 
            spectral_signatures_catalog.

    Examples:
        Fit a classifier
            >>> # Start a session
            >>> import remotior_sensus
            >>> rs = remotior_sensus.Session()
            >>> # create a BandSet
            >>> catalog = rs.bandset_catalog()
            >>> file_list = ['file1.tif', 'file2.tif', 'file3.tif']
            >>> catalog.create_bandset(file_list, wavelengths=['Landsat 8'])
            >>> # set a BandSet reference in signature catalog
            >>> signature_catalog = rs.spectral_signatures_catalog(
            >>>     bandset=catalog.get(1))
            >>> # import vector in signature catalog
            >>> signature_catalog.import_vector(
            >>>     file_path='file.gpkg', macroclass_field='macroclass', class_field='class',
            >>>     macroclass_name_field='macroclass', class_name_field='class',
            >>>     calculate_signature=True)
            >>> # train a minimum distance classifier
            >>> classifier = Classifier.train(
            >>>     spectral_signatures=signature_catalog,
            >>>     algorithm_name='minimum distance'
            >>> )
        """  # noqa: E501

    def __init__(
            self, algorithm_name, spectral_signatures, covariance_matrices,
            model_classifier, input_normalization, normalization_values
    ):
        """Initializes a classifier.

        A classifier is an object which includes the 

        Args:
            algorithm_name: algorithm name selected form cfg.classification_algorithms.
            spectral_signatures: a SpectralSignaturesCatalog containing spectral signatures.
            covariance_matrices: dictionary of previously calculated covariance matrices 
                (used in maximum_likelihood).
            model_classifier: classifier object.
            input_normalization: perform input normalization; options are z_score or linear_scaling.
            normalization_values: list of normalization parameters defined for each variable
                [normalization expressions, mean values, 
                standard deviation values, minimum values, maximum values].
        """  # noqa: E501
        self.algorithm_name = algorithm_name
        self.spectral_signatures = spectral_signatures
        self.covariance_matrices = covariance_matrices
        self.model_classifier = model_classifier
        self.input_normalization = input_normalization
        self.normalization_values = normalization_values
        self.framework_name = None
        self.classification_function = None
        # spectral signatures catalog
        if (algorithm_name == cfg.minimum_distance
                or algorithm_name == cfg.minimum_distance_a
                or algorithm_name == cfg.spectral_angle_mapping
                or algorithm_name == cfg.spectral_angle_mapping_a
                or algorithm_name == cfg.maximum_likelihood
                or algorithm_name == cfg.maximum_likelihood_a):
            if spectral_signatures.signatures is None:
                # import vector
                spectral_signatures.import_vector(
                    file_path=spectral_signatures.geometry_file,
                    macroclass_field=spectral_signatures.macroclass_field,
                    class_field=spectral_signatures.class_field,
                    macroclass_name_field=(
                        spectral_signatures.macroclass_field),
                    class_name_field=spectral_signatures.class_field,
                    calculate_signature=True
                )
            spectral_signatures_catalog = _normalize_signatures(
                spectral_signatures=spectral_signatures,
                normalization_values=normalization_values,
                input_normalization=input_normalization
            )
        else:
            spectral_signatures_catalog = spectral_signatures
        # minimum distance
        if (algorithm_name == cfg.minimum_distance
                or algorithm_name == cfg.minimum_distance_a):
            self.classification_function = classification_minimum_distance
        # spectral angle mapping
        elif (algorithm_name == cfg.spectral_angle_mapping
              or algorithm_name == cfg.spectral_angle_mapping_a):
            self.classification_function = (
                classification_spectral_angle_mapping
            )
        # maximum likelihood
        elif (algorithm_name == cfg.maximum_likelihood
              or algorithm_name == cfg.maximum_likelihood_a):
            self.classification_function = classification_maximum_likelihood
        # scikit framework
        elif (
                algorithm_name == cfg.random_forest
                or algorithm_name == cfg.random_forest_a
                or algorithm_name == cfg.random_forest_ovr
                or algorithm_name == cfg.random_forest_ovr_a
                or algorithm_name == cfg.support_vector_machine
                or algorithm_name == cfg.support_vector_machine_a
                or algorithm_name == cfg.multi_layer_perceptron
                or algorithm_name == cfg.multi_layer_perceptron_a):
            self.classification_function = classification_scikit
            self.framework_name = cfg.scikit_framework
        # pytorch framework
        elif (algorithm_name == cfg.pytorch_multi_layer_perceptron
              or algorithm_name == cfg.pytorch_multi_layer_perceptron_a):
            self.classification_function = classification_pytorch
            self.framework_name = cfg.pytorch_framework
        self.function_argument = {
            cfg.model_classifier_framework: model_classifier,
            cfg.covariance_matrices_framework: covariance_matrices,
            cfg.normalization_values_framework: normalization_values,
            cfg.spectral_signatures_framework: spectral_signatures_catalog
        }

    def save_model(self, output_path: str) -> OutputManager:
        """Saves classifier model.

        Saves a classifier model to file to be loaded later.

        Args:
            output_path: path of output file.
        
        Returns:
            OutputManager object with
                - path = [output path]
            
        Examples:
            Save a trainied classifier
                >>> classifier = Classifier()
                >>> # after the training
                >>> saved_model = classifier.save_model(output_path=output_path)
        """  # noqa: E501
        # list of zipped files
        files = []
        # save classification framework
        file_path = cfg.temp.temporary_file_path(
            name=cfg.classification_framework
        )
        classification_parameters = '%s=%s%s%s=%s' % (
            cfg.algorithm_name_framework, str(self.algorithm_name),
            cfg.new_line, cfg.input_normalization_framework,
            str(self.input_normalization))
        read_write_files.write_file(
            data=classification_parameters, output_path=file_path
        )
        files.append(file_path)
        if self.spectral_signatures is not None:
            file_path = cfg.temp.temporary_file_path(
                name=cfg.spectral_signatures_framework, name_suffix='.pickle'
            )
            # save model using pickle
            with open(file_path, 'wb') as file:
                pickle.dump(self.spectral_signatures, file, protocol=4)
            files.append(file_path)
        if self.normalization_values is not None:
            file_path = cfg.temp.temporary_file_path(
                name=cfg.normalization_values_framework, name_suffix='.pickle'
            )
            # save model using pickle
            with open(file_path, 'wb') as file:
                pickle.dump(self.normalization_values, file, protocol=4)
            files.append(file_path)
        if self.covariance_matrices is not None:
            file_path = cfg.temp.temporary_file_path(
                name=cfg.covariance_matrices_framework, name_suffix='.pickle'
            )
            # save model using pickle
            with open(file_path, 'wb') as file:
                pickle.dump(self.covariance_matrices, file, protocol=4)
            files.append(file_path)
        if self.framework_name == cfg.scikit_framework:
            file_path = cfg.temp.temporary_file_path(
                name=cfg.model_classifier_framework, name_suffix='.pickle'
            )
            # save model using pickle
            with open(file_path, 'wb') as file:
                pickle.dump(self.model_classifier, file, protocol=4)
            files.append(file_path)
        elif self.framework_name == cfg.pytorch_framework:
            file_path = cfg.temp.temporary_file_path(
                name=cfg.model_classifier_framework, name_suffix='.pth'
            )
            torch.save(self.model_classifier, file_path)
            files.append(file_path)
        # zip files
        if files_directories.file_extension(output_path) != cfg.rsmo_suffix:
            output_path = shared_tools.join_path(
                files_directories.parent_directory(output_path), '{}{}'.format(
                    files_directories.file_name(output_path, suffix=False),
                    cfg.rsmo_suffix
                )
            ).replace('\\', '/')
        files_directories.zip_files(files, output_path)
        cfg.logger.log.debug('output_path: %s' % str(output_path))
        return OutputManager(path=output_path)

    @classmethod
    def load_classifier(
            cls, algorithm_name=None, spectral_signatures=None,
            covariance_matrices=None, model_classifier=None,
            input_normalization=None, normalization_values=None
    ):
        """Loads a classifier.

        Creates a classifier from loading.

        Args:
            algorithm_name: algorithm name selected form cfg.classification_algorithms.
            spectral_signatures: a SpectralSignaturesCatalog containing spectral signatures.
            covariance_matrices: dictionary of previously calculated covariance matrices 
                (used in maximum_likelihood).
            model_classifier: classifier object.
            input_normalization: perform input normalization; options are z_score or linear_scaling.
            normalization_values: list of normalization parameters defined for each variable
                [normalization expressions, mean values, 
                standard deviation values, minimum values, maximum values].

        Returns:
            Classifier object.

        Examples:
            Load a classifier
                >>> classifier = Classifier.load_classifier(
                >>> algorithm_name=algorithm_name, spectral_signatures=spectral_signatures,
                >>> covariance_matrices=covariance_matrices, model_classifier=model_classifier,
                >>> input_normalization=input_normalization, normalization_values=normalization_values)
        """  # noqa: E501
        # return classifier
        return cls(
            algorithm_name=algorithm_name,
            spectral_signatures=spectral_signatures,
            covariance_matrices=covariance_matrices,
            model_classifier=model_classifier,
            input_normalization=input_normalization,
            normalization_values=normalization_values
        )

    # noinspection PyTypeChecker
    @classmethod
    def train(
            cls, spectral_signatures=None, algorithm_name=None,
            covariance_matrices=None, svc_classification_confidence=None,
            n_processes: int = None, available_ram: int = None,
            cross_validation=True, x_matrix=None,
            y=None, class_weight=None, input_normalization=None,
            normalization_values=None, find_best_estimator=False,
            rf_max_features=None, rf_number_trees=100,
            rf_min_samples_split=None, svm_c=None, svm_gamma=None,
            svm_kernel=None, pytorch_model=None, pytorch_optimizer=None,
            mlp_training_portion=None, pytorch_loss_function=None,
            mlp_hidden_layer_sizes=None, mlp_alpha=None,
            mlp_learning_rate_init=None, mlp_max_iter=None,
            mlp_batch_size=None, mlp_activation=None,
            pytorch_optimization_n_iter_no_change=None,
            pytorch_optimization_tol=None, pytorch_device=None, min_progress=1,
            max_progress=100
    ):
        """Trains a classifier.

        Trains a classifier using ROIs or spectral signatures.

        Args:
            spectral_signatures: a SpectralSignaturesCatalog containing spectral signatures.
            algorithm_name: algorithm name selected from cfg.classification_algorithms; 
                if None, minimum distance is used.
            n_processes: number of parallel processes.
            available_ram: number of megabytes of RAM available to processes.
            cross_validation: if True, perform cross validation for algorithms
                provided through scikit-learn (random_forest, random_forest_ovr, 
                support_vector_machine, multi_layer_perceptron).
            x_matrix: optional previously saved x matrix.
            y: optional previously saved y matrix.
            covariance_matrices: dictionary of previously calculated covariance matrices 
                (used in maximum_likelihood).
            svc_classification_confidence: if True, write also additional
                classification confidence rasters as output; required information for support_vector_machine.
            input_normalization: perform input normalization; options are z_score or linear_scaling.
            normalization_values: list of normalization paramters defined for each variable
                [normalization expressions, mean values, 
                standar deviation values, minimum values, maximum values].
            class_weight: specific for random forest and support_vector_machine, if None each class
                has equal weight 1, if 'balanced' weight is computed inversely
                proportional to class frequency.
            find_best_estimator: specific for scikit classifiers, if True,
                find automatically the best parameters and fit the model, if
                integer the greater the value the more are the tested combinations.
            rf_max_features: specific for random forest, if None all features
                are considered in node splitting, available options are 'sqrt' as
                square root of all the features, an integer number, or
                a float number for a fraction of all the features.
            rf_number_trees: specific for random forest, number of trees in the forest.
            rf_min_samples_split: specific for random forest through scikit, 
                sets the minimum number of samples required to split an internal node; default = 2.
            svm_c: specific for support_vector_machine through scikit,
                sets the regularization parameter C; default = 1.
            svm_gamma: specific for support_vector_machine through scikit, 
                sets the kernel coefficient; default = scale.
            svm_kernel: specific for support_vector_machine through scikit, 
                sets the kernel; default = rbf.
            mlp_training_portion: specific for multi_layer_perceptron and pytorch_multi_layer_perceptron,
                the proportion of data to be used as training (default = 0.9) and the remaining part
                as test (default = 0.1).
            mlp_hidden_layer_sizes: specific for multi_layer_perceptron and pytorch_multi_layer_perceptron,
                list of values where each value defines the number of neurons in a hidden layer 
                (e.g., [200, 100] for two hidden layers of 200 and 100 neurons respectively); default = [100].
            mlp_alpha: specific for multi_layer_perceptron and pytorch_multi_layer_perceptron,
                weight decay (also L2 regularization term) for Adam optimizer (default = 0.0001).
            mlp_learning_rate_init: specific for multi_layer_perceptron and pytorch_multi_layer_perceptron,
                sets initial learning rate (default = 0.001).
            mlp_max_iter: specific for multi_layer_perceptron and pytorch_multi_layer_perceptron,
                sets the maximum number of iterations (default = 200).
            mlp_batch_size: specific for multi_layer_perceptron and pytorch_multi_layer_perceptron,
                sets the number of samples per batch for optimizer; if "auto", the batch is the 
                minimum value between 200 and the number of samples (default = auto).
            mlp_activation: specific for multi_layer_perceptron and pytorch_multi_layer_perceptron,
                sets the activation function (default relu).
            pytorch_model: specific for pytorch_multi_layer_perceptron,
                custom pytorch nn.Module.
            pytorch_optimizer: specific for pytorch_multi_layer_perceptron,
                custom pytorch optimizer.
            pytorch_loss_function: specific for pytorch_multi_layer_perceptron,
                sets a custom loss function (default CrossEntropyLoss).
            pytorch_optimization_n_iter_no_change: specific for pytorch_multi_layer_perceptron,
                sets the maximum number of epochs where the loss is not improving by 
                at least the value pytorch_optimization_tol (default 5).
            pytorch_optimization_tol: specific for pytorch_multi_layer_perceptron,
                sets the tolerance of optimization (default = 0.0001).
            pytorch_device: specific for pytorch_multi_layer_perceptron,
                processing device 'cpu' (default) or 'cuda' if available.
            min_progress: minimum progress value for :func:`~remotior_sensus.core.progress.Progress`.
            max_progress: maximum progress value for :func:`~remotior_sensus.core.progress.Progress`.

        Returns:
            Classifier object.

        Examples:
            Load a classifier
                >>> import remotior_sensus
                >>> rs = remotior_sensus.Session()
                >>> signature_catalog = rs.spectral_signatures_catalog()
                >>> classifier = Classifier.train(
                >>> spectral_signatures=spectral_signatures,
                >>> algorithm_name='minimum distance')
        """  # noqa: E501
        cfg.logger.log.info('start')
        # classification parameters
        model_classifier = None
        if n_processes is None:
            n_processes = cfg.n_processes
        if available_ram is None:
            available_ram = cfg.available_ram
        # random forest scikit
        if (algorithm_name == cfg.random_forest
                or algorithm_name == cfg.random_forest_a):
            if (find_best_estimator is not None
                    and find_best_estimator is not False):
                if type(find_best_estimator) is int:
                    n_steps = find_best_estimator
                else:
                    n_steps = 5
                if rf_max_features is None:
                    list_rf_max_features = [0.3, 0.65, 1]
                else:
                    list_rf_max_features = [rf_max_features]
                # calculate grid steps
                n_estimators_step = int((rf_number_trees - 10) / n_steps)
                if n_estimators_step <= 0:
                    n_estimators_step = 1
                min_samples_split_step = round((10 - 1) / n_steps)
                if min_samples_split_step <= 0:
                    min_samples_split_step = 1
                # build function argument list of dictionaries
                argument_list = []
                function_list = []
                for n_estimators in range(
                        10, rf_number_trees, n_estimators_step
                ):
                    for min_samples_split in range(
                            2, 10,
                            min_samples_split_step
                    ):
                        for max_features in list_rf_max_features:
                            for class_weight in [None, 'balanced']:
                                argument_list.append(
                                    {
                                        'classifier': RandomForestClassifier(
                                            criterion='gini',
                                            n_estimators=n_estimators,
                                            oob_score=False,
                                            max_features=max_features,
                                            class_weight=class_weight,
                                            n_jobs=1, min_samples_split=(
                                                min_samples_split),
                                            verbose=0, random_state=0
                                        ), 'x_matrix': x_matrix, 'y': y
                                    }
                                )
                                function_list.append(
                                    score_classifier_stratified
                                )
                cfg.multiprocess.run_iterative_process(
                    function_list=function_list, argument_list=argument_list,
                    min_progress=min_progress, max_progress=max_progress
                )
                if cfg.multiprocess.output is False:
                    cfg.logger.log.error('unable to calculate')
                    cfg.messages.error('unable to calculate')
                    return cls(
                        algorithm_name=None,
                        spectral_signatures=None,
                        covariance_matrices=None,
                        model_classifier=None,
                        input_normalization=None,
                        normalization_values=None
                    )
                results = cfg.multiprocess.output
                # fit classifiers
                score_list = []
                score_std_list = []
                classifier_list = []
                for result in results:
                    for r in result:
                        classifier_list.append(r[0])
                        score_list.append(r[1])
                        score_std_list.append(r[2])
                score_index = score_list.index(max(score_list))
                cfg.logger.log.debug(
                    'cross validation score max: %s; score std: %s' % (
                        str(max(score_list)), str(score_std_list[score_index]))
                )
                model_classifier = classifier_list[score_index]
                model_classifier.oob_score = True
                model_classifier.n_jobs = n_processes
                cfg.progress.update(message='fitting')
                model_classifier.fit(x_matrix, y)
                cfg.logger.log.info(
                    'best parameters: %s; feature importance: %s; '
                    'accuracy score: %s' % (
                        str(model_classifier.get_params()), str(
                            model_classifier.feature_importances_
                        ), str(model_classifier.oob_score_))
                )
            else:
                if rf_min_samples_split is None:
                    rf_min_samples_split = 2
                cfg.logger.log.debug(
                    'n_estimators: %s; max_features: %s; class_weight: %s; '
                    'min_samples_split: %s' % (
                        str(rf_number_trees), str(rf_max_features),
                        str(class_weight), str(rf_min_samples_split))
                )
                # build classifier
                try:
                    model_classifier = RandomForestClassifier(
                        criterion='gini', n_estimators=rf_number_trees,
                        oob_score=False, max_features=rf_max_features,
                        class_weight=class_weight, n_jobs=1,
                        min_samples_split=rf_min_samples_split, verbose=2,
                        random_state=0
                    )
                except Exception as err:
                    cfg.logger.log.error(str(err))
                    cfg.messages.error(str(err))
                    return cls(
                        algorithm_name=None,
                        spectral_signatures=None,
                        covariance_matrices=None,
                        model_classifier=None,
                        input_normalization=None,
                        normalization_values=None
                    )
                # perform cross validation if cross_validation is True
                _perform_cross_validation_scikit(
                    cross_validation=cross_validation,
                    classifier=model_classifier, x_matrix=x_matrix, y=y,
                    n_processes=n_processes, available_ram=available_ram
                )
                # fit classifier
                model_classifier.oob_score = True
                model_classifier.n_jobs = n_processes
                cfg.progress.update(message='fitting')
                model_classifier.fit(x_matrix, y)
                model_classifier.n_jobs = 1
                cfg.logger.log.info(
                    'feature importance: %s; accuracy score: %s' % (
                        str(model_classifier.feature_importances_),
                        str(model_classifier.oob_score_))
                )
        # random forest ovr scikit
        elif (algorithm_name == cfg.random_forest_ovr
              or algorithm_name == cfg.random_forest_ovr_a):
            if (
                    find_best_estimator is not None and find_best_estimator
                    is not False):
                if type(find_best_estimator) is int:
                    n_steps = find_best_estimator
                else:
                    n_steps = 5
                if rf_max_features is None:
                    list_rf_max_features = [0.3, 0.65, 1]
                else:
                    list_rf_max_features = [rf_max_features]
                # calculate grid steps
                n_estimators_step = int((rf_number_trees - 10) / n_steps)
                if n_estimators_step <= 0:
                    n_estimators_step = 1
                min_samples_split_step = round((10 - 1) / n_steps)
                if min_samples_split_step <= 0:
                    min_samples_split_step = 1
                # build function argument list of dictionaries
                argument_list = []
                function_list = []
                for n_estimators in range(
                        10, rf_number_trees, n_estimators_step
                ):
                    for min_samples_split in range(
                            2, 10, min_samples_split_step
                    ):
                        for max_features in list_rf_max_features:
                            for class_weight in [None, 'balanced']:
                                argument_list.append(
                                    {
                                        'classifier': OneVsRestClassifier(
                                            RandomForestClassifier(
                                                criterion='gini',
                                                n_estimators=n_estimators,
                                                oob_score=False,
                                                max_features=max_features,
                                                class_weight=class_weight,
                                                n_jobs=1,
                                                min_samples_split=(
                                                    min_samples_split),
                                                verbose=0, random_state=0
                                            )
                                        ), 'x_matrix': x_matrix, 'y': y
                                    }
                                )
                                function_list.append(
                                    score_classifier_stratified
                                )
                cfg.multiprocess.run_iterative_process(
                    function_list=function_list, argument_list=argument_list,
                    min_progress=min_progress, max_progress=max_progress
                )
                if cfg.multiprocess.output is False:
                    cfg.logger.log.error('unable to calculate')
                    cfg.messages.error('unable to calculate')
                    return cls(
                        algorithm_name=None,
                        spectral_signatures=None,
                        covariance_matrices=None,
                        model_classifier=None,
                        input_normalization=None,
                        normalization_values=None
                    )
                results = cfg.multiprocess.output
                # fit classifiers
                score_list = []
                score_std_list = []
                classifier_list = []
                for result in results:
                    for r in result:
                        classifier_list.append(r[0])
                        score_list.append(r[1])
                        score_std_list.append(r[2])
                score_index = score_list.index(max(score_list))
                model_classifier = classifier_list[score_index]
                cfg.progress.update(message='fitting')
                model_classifier.fit(x_matrix, y)
                cfg.logger.log.info(
                    'best parameters: %s; cross validation score max: %s; '
                    'score std: %s' % (str(model_classifier.get_params()),
                                       str(max(score_list)),
                                       str(score_std_list[score_index]))
                )
            else:
                if rf_min_samples_split is None:
                    rf_min_samples_split = 2
                cfg.logger.log.debug(
                    'n_estimators: %s; max_features: %s; class_weight: %s; '
                    'min_samples_split: %s' % (
                        str(rf_number_trees), str(rf_max_features),
                        str(class_weight), str(rf_min_samples_split))
                )
                # build classifier
                model_classifier = OneVsRestClassifier(
                    RandomForestClassifier(
                        criterion='gini', n_estimators=rf_number_trees,
                        oob_score=True, max_features=rf_max_features,
                        class_weight=class_weight, n_jobs=1,
                        min_samples_split=rf_min_samples_split, random_state=0
                    )
                )
                # perform cross validation if cross_validation is True
                _perform_cross_validation_scikit(
                    cross_validation=cross_validation,
                    classifier=model_classifier, x_matrix=x_matrix, y=y,
                    n_processes=n_processes, available_ram=available_ram
                )
                # fit classifier
                cfg.progress.update(message='fitting')
                model_classifier.fit(x_matrix, y)
                cfg.logger.log.info(
                    'OneVsRestClassifier: RandomForestClassifier'
                )
        # support vector machine scikit
        elif (algorithm_name == cfg.support_vector_machine
              or algorithm_name == cfg.support_vector_machine_a):
            if svc_classification_confidence is None:
                svc_classification_confidence = False
            if (
                    find_best_estimator is not None and find_best_estimator
                    is not False):
                if type(find_best_estimator) is int:
                    n_steps = find_best_estimator
                else:
                    n_steps = 5
                # calculate grid steps
                kernels = ['linear', 'rbf']
                c_steps = np.logspace(-5, 10, n_steps).tolist()
                gamma_steps = np.logspace(-10, 3, n_steps).tolist()
                gamma_steps.append('scale')
                # build function argument list of dictionaries
                argument_list = []
                function_list = []
                for kernel in kernels:
                    for c_step in c_steps:
                        for gamma in gamma_steps:
                            for class_weight in [None, 'balanced']:
                                argument_list.append(
                                    {
                                        'classifier': svm.SVC(
                                            probability=(
                                                svc_classification_confidence),
                                            kernel=kernel, gamma=gamma,
                                            C=c_step,
                                            class_weight=class_weight,
                                            cache_size=int(
                                                available_ram / (
                                                        2 * n_processes)
                                            ), verbose=0, random_state=0
                                        ), 'x_matrix': x_matrix, 'y': y
                                    }
                                )
                                function_list.append(
                                    score_classifier_stratified
                                )
                cfg.multiprocess.run_iterative_process(
                    function_list=function_list, argument_list=argument_list,
                    min_progress=min_progress, max_progress=max_progress
                )
                if cfg.multiprocess.output is False:
                    cfg.logger.log.error('unable to calculate')
                    cfg.messages.error('unable to calculate')
                    return cls(
                        algorithm_name=None,
                        spectral_signatures=None,
                        covariance_matrices=None,
                        model_classifier=None,
                        input_normalization=None,
                        normalization_values=None
                    )
                results = cfg.multiprocess.output
                # fit classifier
                score_list = []
                score_std_list = []
                classifier_list = []
                for result in results:
                    for r in result:
                        classifier_list.append(r[0])
                        score_list.append(r[1])
                        score_std_list.append(r[2])
                score_index = score_list.index(max(score_list))
                model_classifier = classifier_list[score_index]
                cfg.progress.update(message='fitting')
                model_classifier.fit(x_matrix, y)
                cfg.logger.log.info(
                    'best parameters: %s; cross validation score max: %s; '
                    'score std: %s' % (str(model_classifier.get_params()),
                                       str(max(score_list)),
                                       str(score_std_list[score_index]))
                )
            else:
                if svm_c is None:
                    svm_c = 1
                if svm_gamma is None:
                    svm_gamma = 'scale'
                if svm_kernel is None:
                    svm_kernel = 'rbf'
                cfg.logger.log.debug(
                    'kernel: %s; gamma: %s; c: %s; class_weight: %s' % (
                        str(svm_kernel), str(svm_gamma), str(svm_c),
                        str(class_weight))
                )
                # build classifier
                model_classifier = svm.SVC(
                    probability=svc_classification_confidence,
                    kernel=svm_kernel, gamma=svm_gamma, C=svm_c,
                    random_state=0, class_weight=class_weight,
                    cache_size=int(available_ram / (2 * n_processes)),
                    verbose=0
                )
                # perform cross validation if cross_validation is True
                _perform_cross_validation_scikit(
                    cross_validation=cross_validation,
                    classifier=model_classifier, x_matrix=x_matrix, y=y,
                    n_processes=n_processes, available_ram=available_ram
                )
                # fit classifier
                model_classifier.cache_size = int(available_ram / 3)
                cfg.progress.update(message='fitting')
                model_classifier.fit(x_matrix, y)
        # multilayer perceptron scikit
        elif (algorithm_name == cfg.multi_layer_perceptron
              or algorithm_name == cfg.multi_layer_perceptron_a):
            if mlp_training_portion is None:
                mlp_training_portion = 0.9
            elif mlp_training_portion > 1:
                mlp_training_portion = 0.9
            if mlp_activation is None:
                mlp_activation = 'relu'
            if (
                    find_best_estimator is not None and find_best_estimator
                    is not False):
                if type(find_best_estimator) is int:
                    n_steps = find_best_estimator
                else:
                    n_steps = 5
                if mlp_max_iter is None:
                    mlp_max_iter = 200
                # calculate grid steps
                hidden_layer_sizes_steps = list(range(50, 500, 100))
                if n_steps > 1:
                    for neurons_1 in range(50, 200, 50):
                        for neurons_2 in range(50, 200, 50):
                            hidden_layer_sizes_steps.append(
                                [neurons_1, neurons_2]
                            )
                if n_steps > 3:
                    for neurons_1 in range(50, 150, 50):
                        for neurons_2 in range(50, 150, 50):
                            for neurons_3 in range(50, 150, 50):
                                hidden_layer_sizes_steps.append(
                                    [neurons_1, neurons_2, neurons_3]
                                )
                if n_steps > 5:
                    for neurons_1 in range(25, 75, 25):
                        for neurons_2 in range(25, 75, 25):
                            for neurons_3 in range(25, 75, 25):
                                for neurons_4 in range(25, 75, 25):
                                    hidden_layer_sizes_steps.append(
                                        [neurons_1, neurons_2, neurons_3,
                                         neurons_4]
                                    )
                learning_rate_init_steps = np.logspace(
                    -4, -1, n_steps
                ).tolist()
                alpha_steps = np.logspace(-5, -2, n_steps).tolist()
                batch_max = max(200, int(x_matrix.shape[0] / 100))
                batch_min = int(batch_max / n_steps)
                batch_size_steps = [batch_min, batch_max]
                # build function argument list of dictionaries
                argument_list = []
                function_list = []
                for alpha in alpha_steps:
                    for learning_rate_init in learning_rate_init_steps:
                        for hidden_layer_sizes in hidden_layer_sizes_steps:
                            for batch in batch_size_steps:
                                argument_list.append(
                                    {
                                        'classifier': MLPClassifier(
                                            hidden_layer_sizes=(
                                                hidden_layer_sizes),
                                            alpha=alpha,
                                            learning_rate_init=(
                                                learning_rate_init),
                                            validation_fraction=(
                                                    1 - mlp_training_portion),
                                            activation=mlp_activation,
                                            random_state=0, batch_size=batch,
                                            max_iter=mlp_max_iter, verbose=0
                                        ), 'x_matrix': x_matrix, 'y': y
                                    }
                                )
                                function_list.append(
                                    score_classifier_stratified
                                )
                cfg.multiprocess.run_iterative_process(
                    function_list=function_list, argument_list=argument_list,
                    min_progress=min_progress, max_progress=max_progress
                )
                if cfg.multiprocess.output is False:
                    cfg.logger.log.error('unable to calculate')
                    cfg.messages.error('unable to calculate')
                    return cls(
                        algorithm_name=None,
                        spectral_signatures=None,
                        covariance_matrices=None,
                        model_classifier=None,
                        input_normalization=None,
                        normalization_values=None
                    )
                results = cfg.multiprocess.output
                # fit classifier
                score_list = []
                score_std_list = []
                classifier_list = []
                for result in results:
                    for r in result:
                        classifier_list.append(r[0])
                        score_list.append(r[1])
                        score_std_list.append(r[2])
                score_index = score_list.index(max(score_list))
                model_classifier = classifier_list[score_index]
                cfg.progress.update(message='fitting')
                model_classifier.fit(x_matrix, y)
                cfg.logger.log.info(
                    'best parameters: %s; cross validation score max: %s; '
                    'score std: %s' % (str(model_classifier.get_params()),
                                       str(max(score_list)),
                                       str(score_std_list[score_index]))
                )
            else:
                if mlp_hidden_layer_sizes is None:
                    mlp_hidden_layer_sizes = [100]
                if mlp_alpha is None:
                    mlp_alpha = 0.0001
                if mlp_learning_rate_init is None:
                    mlp_learning_rate_init = 0.001
                if mlp_max_iter is None:
                    mlp_max_iter = 200
                if mlp_batch_size is None:
                    mlp_batch_size = int(
                        min(
                            2000, min(
                                x_matrix.shape[0], max(
                                    200, x_matrix.shape[0] / 100
                                )
                            )
                        )
                    )
                cfg.logger.log.debug(
                    'hidden_layer_sizes: %s; alpha: %s; learning_rate_init: '
                    '%s; max_iter: %s; batch_size: %s;  activation: %s' % (
                        str(mlp_hidden_layer_sizes), str(mlp_alpha),
                        str(mlp_learning_rate_init), str(mlp_max_iter),
                        str(mlp_batch_size), str(mlp_activation))
                )
                # build classifier
                model_classifier = MLPClassifier(
                    hidden_layer_sizes=mlp_hidden_layer_sizes, alpha=mlp_alpha,
                    activation=mlp_activation,
                    learning_rate_init=mlp_learning_rate_init,
                    max_iter=mlp_max_iter, batch_size=mlp_batch_size,
                    verbose=0, random_state=0,
                    validation_fraction=1 - mlp_training_portion
                )
                # perform cross validation if cross_validation is True
                _perform_cross_validation_scikit(
                    cross_validation=cross_validation,
                    classifier=model_classifier, x_matrix=x_matrix, y=y,
                    n_processes=n_processes, available_ram=available_ram
                )
                # fit classifier
                cfg.progress.update(message='fitting')
                model_classifier.fit(x_matrix, y)
        # multilayer perceptron pytorch
        elif (algorithm_name == cfg.pytorch_multi_layer_perceptron
              or algorithm_name == cfg.pytorch_multi_layer_perceptron_a):
            if mlp_training_portion is None:
                mlp_training_portion = 0.9
            elif mlp_training_portion > 1:
                mlp_training_portion = 0.9
            if mlp_max_iter is None:
                mlp_max_iter = 200
            if mlp_batch_size is None or mlp_batch_size == 'auto':
                mlp_batch_size = int(
                    min(
                        2000, min(
                            x_matrix.shape[0],
                            max(200, x_matrix.shape[0] / 100)
                        )
                    )
                )
            cfg.progress.update(message='fitting')
            (model_classifier, training_loss, test_loss,
             accuracy) = train_pytorch_model(
                x_matrix=x_matrix, y_matrix=y, pytorch_model=pytorch_model,
                activation=mlp_activation, batch_size=mlp_batch_size,
                n_processes=n_processes, training_portion=mlp_training_portion,
                pytorch_optimizer=pytorch_optimizer,
                hidden_layer_sizes=mlp_hidden_layer_sizes,
                loss_function=pytorch_loss_function,
                learning_rate_init=mlp_learning_rate_init,
                optimization_n_iter_no_change=(
                    pytorch_optimization_n_iter_no_change
                ), optimization_tol=pytorch_optimization_tol,
                weight_decay=mlp_alpha, max_iterations=mlp_max_iter,
                device=pytorch_device, min_progress=min_progress,
                max_progress=max_progress
            )
        cfg.logger.log.info('end')
        # return classifier
        return cls(
            algorithm_name=algorithm_name,
            spectral_signatures=spectral_signatures,
            covariance_matrices=covariance_matrices,
            model_classifier=model_classifier,
            input_normalization=input_normalization,
            normalization_values=normalization_values
        )

    def run_prediction(
            self, input_raster_list, output_raster_path,
            n_processes: Optional[int] = None,
            available_ram: Optional[int] = None,
            macroclass: Optional[bool] = True,
            threshold: Optional[bool] = False,
            signature_raster: Optional[bool] = False,
            classification_confidence: Optional[bool] = False,
            virtual_raster: Optional[bool] = None,
            min_progress: Optional[int] = 1, max_progress: Optional[int] = 100
    ):
        """Runs prediction.

        Performs multiprocess classification using a trained classifier using 
        input bands.

        Args:
            input_raster_list: list of paths of input rasters.
            output_raster_path: path of output file.
            n_processes: number of parallel processes.
            available_ram: number of megabytes of RAM available to processes.
            macroclass: if True, use macroclass ID from ROIs or spectral signatures; if False use class ID.
            threshold: if False, classification without threshold; if True,
                use single threshold for each signature; if float, use this value 
                as threshold for all the signature.
            classification_confidence: if True, write also additional
                classification confidence rasters as output.
            signature_raster: if True, write additional rasters for each
                spectral signature as output.
            virtual_raster: if True, create virtual raster output.
            min_progress: minimum progress value for :func:`~remotior_sensus.core.progress.Progress`.
            max_progress: maximum progress value for :func:`~remotior_sensus.core.progress.Progress`.


        Returns:
            OutputManager object with
                - path = [output path]

        Examples:
            Save a trainied classifier
                >>> classifier = Classifier()
                >>> # after the training
                >>> prediction = classifier.run_prediction(
                ... input_raster_list=['file1.tif', 'file2.tif', 'file3.tif'], 
                ... output_raster_path='file_path')
        """  # noqa: E501
        prepared = shared_tools.prepare_input_list(input_raster_list)
        same_geotransformation = prepared['same_geotransformation']
        if same_geotransformation is True:
            vrt_check = input_raster_list
        else:
            # create virtual raster of input
            vrt_check = raster_vector.create_temporary_virtual_raster(
                input_raster_list
            )
        if n_processes is None:
            n_processes = cfg.n_processes
        # dummy bands for memory calculation
        dummy_bands = 5
        if self.classification_function is not None:
            cfg.multiprocess.run(
                raster_path=vrt_check, function=self.classification_function,
                function_argument=self.function_argument,
                n_processes=n_processes,
                available_ram=available_ram, dummy_bands=dummy_bands,
                function_variable=[macroclass, threshold],
                output_raster_path=output_raster_path, classification=True,
                classification_confidence=classification_confidence,
                signature_raster=signature_raster,
                virtual_raster=virtual_raster,
                progress_message='classification', min_progress=min_progress,
                max_progress=max_progress
            )
            cfg.logger.log.debug('output_path: %s' % str(output_raster_path))
            return OutputManager(path=output_raster_path)
        else:
            cfg.logger.log.error('classification function not available')
            return OutputManager(check=False)


def band_classification(
        input_bands: Union[list, int, BandSet],
        output_path: Optional[str] = None, overwrite: Optional[bool] = False,
        spectral_signatures: Optional[
            Union[str, SpectralSignaturesCatalog]] = None,
        algorithm_name: Optional[str] = None,
        bandset_catalog: Optional[BandSetCatalog] = None,
        macroclass: Optional[bool] = True,
        threshold: Optional[Union[bool, float]] = False,
        classification_confidence: Optional[bool] = False,
        signature_raster: Optional[bool] = False,
        n_processes: Optional[int] = None,
        available_ram: Optional[int] = None,
        cross_validation: Optional[bool] = True,
        x_input: Optional[np.array] = None, y_input: Optional[np.array] = None,
        covariance_matrices: Optional[dict] = None,
        input_normalization: Optional[str] = None,
        load_classifier: Optional[str] = None,
        save_classifier: Optional[bool] = False,
        only_fit: Optional[bool] = False,
        class_weight: Optional[Union[None, str, dict]] = None,
        find_best_estimator=False, rf_max_features=None,
        rf_number_trees: Optional[int] = 100,
        rf_min_samples_split: Optional[Union[None, int, float]] = None,
        svm_c: Optional[float] = None,
        svm_gamma: Optional[Union[str, float]] = None,
        svm_kernel: Optional[str] = None,
        mlp_training_portion: Optional[Union[None, float]] = None,
        mlp_hidden_layer_sizes: Optional[Union[None, tuple, list]] = None,
        mlp_alpha: Optional[float] = None,
        mlp_learning_rate_init: Optional[float] = None,
        mlp_max_iter: Optional[float] = None,
        mlp_batch_size: Optional[Union[None, int, str]] = None,
        mlp_activation: Optional[Union[None, str]] = None,
        pytorch_model: Optional = None, pytorch_optimizer: Optional = None,
        pytorch_loss_function: Optional = None,
        pytorch_optimization_n_iter_no_change: Optional[
            Union[None, int]] = None,
        pytorch_optimization_tol: Optional[Union[None, int]] = None,
        pytorch_device: Optional[Union[None, str]] = None,
        progress_message: Optional[bool] = True
) -> OutputManager:
    """Performs band classification.

    This tool allows for classification of raster bands using the selected
    algorithm.

    Args:
        input_bands: list of input raster paths, or a BandSet number,
            or a previously defined BandSet.
        output_path: path of output file.
        overwrite: if True, output overwrites existing files.
        spectral_signatures: a SpectralSignaturesCatalog containing spectral signatures.
        algorithm_name: algorithm name selected from cfg.classification_algorithms; 
            if None, minimum distance is used.
        bandset_catalog: BandSetCatalog object.
        macroclass: if True, use macroclass ID from ROIs or spectral signatures; if False use class ID.
        threshold: if False, classification without threshold; if True,
            use single threshold for each signature; if float, use this value 
            as threshold for all the signature.
        classification_confidence: if True, write also additional
            classification confidence rasters as output.
        signature_raster: if True, write additional rasters for each
            spectral signature as output.
        n_processes: number of parallel processes.
        available_ram: number of megabytes of RAM available to processes.
        cross_validation: if True, perform cross validation for algorithms
            provided through scikit-learn (random_forest, random_forest_ovr, 
            support_vector_machine, multi_layer_perceptron).
        load_classifier: path to a previously saved classifier.
        x_input: optional previously saved x matrix.
        y_input: optional previously saved y matrix.
        covariance_matrices: dictionary of previously calculated covariance matrices 
            (used in maximum_likelihood).
        input_normalization: perform input normalization; options are z_score or linear_scaling.
        only_fit: perform only classifier fitting.
        save_classifier: save classifier to file.
        class_weight: specific for random forest and support_vector_machine, if None each class
            has equal weight 1, if 'balanced' weight is computed inversely
            proportional to class frequency.
        find_best_estimator: specific for scikit classifiers, if True,
            find automatically the best parameters and fit the model, if
            integer the greater the value the more are the tested combinations.
        rf_max_features: specific for random forest, if None all features
            are considered in node splitting, available options are 'sqrt' as
            square root of all the features, an integer number, or
            a float number for a fraction of all the features.
        rf_number_trees: specific for random forest, number of trees in the forest.
        rf_min_samples_split: specific for random forest through scikit, 
            sets the minimum number of samples required to split an internal node; default = 2.
        svm_c: specific for support_vector_machine through scikit,
            sets the regularization parameter C; default = 1.
        svm_gamma: specific for support_vector_machine through scikit, 
            sets the kernel coefficient; default = scale.
        svm_kernel: specific for support_vector_machine through scikit, 
            sets the kernel; default = rbf.
        mlp_training_portion: specific for multi_layer_perceptron and pytorch_multi_layer_perceptron,
            the proportion of data to be used as training (default = 0.9) and the remaining part
            as test (default = 0.1).
        mlp_hidden_layer_sizes: specific for multi_layer_perceptron and pytorch_multi_layer_perceptron,
            list of values where each value defines the number of neurons in a hidden layer 
            (e.g., [200, 100] for two hidden layers of 200 and 100 neurons respectively); default = [100].
        mlp_alpha: specific for multi_layer_perceptron and pytorch_multi_layer_perceptron,
            weight decay (also L2 regularization term) for Adam optimizer (default = 0.0001).
        mlp_learning_rate_init: specific for multi_layer_perceptron and pytorch_multi_layer_perceptron,
            sets initial learning rate (default = 0.001).
        mlp_max_iter: specific for multi_layer_perceptron and pytorch_multi_layer_perceptron,
            sets the maximum number of iterations (default = 200).
        mlp_batch_size: specific for multi_layer_perceptron and pytorch_multi_layer_perceptron,
            sets the number of samples per batch for optimizer; if "auto", the batch is the 
            minimum value between 200 and the number of samples (default = auto).
        mlp_activation: specific for multi_layer_perceptron and pytorch_multi_layer_perceptron,
            sets the activation function (default relu).
        pytorch_model: specific for pytorch_multi_layer_perceptron,
            custom pytorch nn.Module.
        pytorch_optimizer: specific for pytorch_multi_layer_perceptron,
            custom pytorch optimizer.
        pytorch_loss_function: specific for pytorch_multi_layer_perceptron,
            sets a custom loss function (default CrossEntropyLoss).
        pytorch_optimization_n_iter_no_change: specific for pytorch_multi_layer_perceptron,
            sets the maximum number of epochs where the loss is not improving by 
            at least the value pytorch_optimization_tol (default 5).
        pytorch_optimization_tol: specific for pytorch_multi_layer_perceptron,
            sets the tolerance of optimization (default = 0.0001).
        pytorch_device: specific for pytorch_multi_layer_perceptron,
            processing device 'cpu' (default) or 'cuda' if available.
        progress_message: progress message.

    Returns:
        If only_fit is True returns :func:`~remotior_sensus.core.output_manager.OutputManager` object with
            - extra = {'classifier': classifier, 'model_path': output model path}

        If only_fit is False returns :func:`~remotior_sensus.core.output_manager.OutputManager` object with
            - path = classification path
            - extra = {'model_path': output model path}
    """  # noqa: E501
    cfg.logger.log.info('start')
    if progress_message:
        cfg.progress.update(
            process=__name__.split('.')[-1].replace('_', ' '),
            message='starting', start=True
        )
    if algorithm_name is None:
        algorithm_name = cfg.minimum_distance
    elif algorithm_name not in cfg.classification_algorithms:
        cfg.logger.log.error('unknown algorithm name')
        cfg.messages.error('unknown algorithm name')
        return OutputManager(check=False)
    cfg.progress.update(message='starting the classifier', step=1)
    cfg.logger.log.debug('algorithm_name: %s' % str(algorithm_name))
    # prepare process files
    prepared = shared_tools.prepare_process_files(
        input_bands=input_bands, output_path=output_path, overwrite=overwrite,
        n_processes=n_processes, bandset_catalog=bandset_catalog,
        multiple_input=True
    )
    input_raster_list = prepared['input_raster_list']
    out_path = prepared['output_path']
    vrt_r = prepared['virtual_output']
    n_processes = prepared['n_processes']
    if load_classifier is not None:
        loaded_classifier = _load_model(load_classifier)
        if loaded_classifier.check:
            classifier = loaded_classifier.extra['classifier']
        else:
            cfg.logger.log.error('failed loading classifier')
            cfg.messages.error('failed loading classifier')
            return OutputManager(check=False)
    else:
        # get spectral signature file
        if type(spectral_signatures) is str:
            if type(input_bands) is BandSet:
                bandset_signature = input_bands
            else:
                bandset_signature = None
            signature_catalog = SpectralSignaturesCatalog(
                bandset=bandset_signature
            )
            signature_catalog.load(file_path=spectral_signatures)
            spectral_signatures = signature_catalog
        # collect x and y matrices
        x_y_matrices = _collect_x_y_matrices(
            x_matrix=x_input, y=y_input,
            covariance_matrices=covariance_matrices,
            input_raster_list=input_raster_list,
            spectral_signatures=spectral_signatures,
            input_normalization=input_normalization, macroclass=macroclass,
            n_processes=n_processes, available_ram=available_ram,
            min_progress=1, max_progress=5, algorithm_name=algorithm_name
        )
        if x_y_matrices.check is False:
            return OutputManager(check=False)
        x_matrix = x_y_matrices.extra['x']
        y = x_y_matrices.extra['y']
        covariance_matrices_dict = x_y_matrices.extra['covariance_matrices']
        normalization = x_y_matrices.extra['normalization_values']
        # fit classifier
        classifier = Classifier.train(
            spectral_signatures=spectral_signatures,
            algorithm_name=algorithm_name,
            covariance_matrices=covariance_matrices_dict,
            svc_classification_confidence=classification_confidence,
            n_processes=n_processes, available_ram=available_ram,
            cross_validation=cross_validation, x_matrix=x_matrix, y=y,
            class_weight=class_weight, input_normalization=input_normalization,
            normalization_values=normalization,
            find_best_estimator=find_best_estimator,
            rf_max_features=rf_max_features, rf_number_trees=rf_number_trees,
            rf_min_samples_split=rf_min_samples_split, svm_c=svm_c,
            svm_gamma=svm_gamma, svm_kernel=svm_kernel,
            pytorch_model=pytorch_model, pytorch_optimizer=pytorch_optimizer,
            mlp_training_portion=mlp_training_portion,
            pytorch_loss_function=pytorch_loss_function,
            mlp_hidden_layer_sizes=mlp_hidden_layer_sizes, mlp_alpha=mlp_alpha,
            mlp_learning_rate_init=mlp_learning_rate_init,
            mlp_max_iter=mlp_max_iter, mlp_batch_size=mlp_batch_size,
            mlp_activation=mlp_activation,
            pytorch_optimization_n_iter_no_change=(
                pytorch_optimization_n_iter_no_change),
            pytorch_optimization_tol=pytorch_optimization_tol,
            pytorch_device=pytorch_device, min_progress=5, max_progress=40
        )
    # save model
    if not save_classifier:
        output_model = None
    else:
        saved_model = classifier.save_model(output_path=output_path)
        output_model = saved_model.path
    # perform multiprocess classification
    if not only_fit:
        prediction = classifier.run_prediction(
            input_raster_list=input_raster_list, output_raster_path=out_path,
            n_processes=n_processes, available_ram=available_ram,
            macroclass=macroclass,
            threshold=threshold, signature_raster=signature_raster,
            classification_confidence=classification_confidence,
            virtual_raster=vrt_r, min_progress=40, max_progress=90
        )
        if prediction.check:
            cfg.logger.log.info('end; prediction: %s' % str(prediction.path))
            return OutputManager(
                path=prediction.path, extra={'model_path': output_model}
            )
    else:
        cfg.logger.log.info(
            'end; only_fit: %s; output_model: %s' % (
                str(only_fit), str(output_model))
        )
        return OutputManager(
            extra={'classifier': classifier, 'model_path': output_model}
        )


def _perform_cross_validation_scikit(
        cross_validation, classifier, x_matrix, y, n_processes: int = None,
        available_ram: int = None
):
    """Performs cross validation with scikit."""
    if cross_validation and cross_validation is not None:
        # calculate score by cross-validation with stratification
        classifier_list, train_arg_dict_list, test_arg_dict_list = \
            _create_stratified_k_fold_scikit(
                classifier=classifier, x_matrix=x_matrix, y=y
            )
        # fit training
        cfg.multiprocess.run_scikit(
            function=fit_classifier, classifier_list=classifier_list,
            list_train_argument_dictionaries=train_arg_dict_list,
            min_progress=5, max_progress=40, n_processes=n_processes,
            available_ram=available_ram
        )
        if cfg.multiprocess.output is False:
            cfg.logger.log.error('unable to calculate')
            cfg.messages.error('unable to calculate')
            return False
        classifiers_fitted = cfg.multiprocess.output
        fitted_classifier_list = shared_tools.expand_list(classifiers_fitted)
        # score test
        cfg.multiprocess.run_scikit(
            function=score_classifier, classifier_list=fitted_classifier_list,
            list_train_argument_dictionaries=test_arg_dict_list,
            n_processes=n_processes, available_ram=available_ram
        )
        if cfg.multiprocess.output is False:
            cfg.logger.log.error('unable to calculate')
            cfg.messages.error('unable to calculate')
            return False
        test_scores = cfg.multiprocess.output
        score_list = shared_tools.expand_list(test_scores)
        scores = np.array(score_list)
        cfg.logger.log.debug(
            'cross validation score mean: %s; cross validation score std: '
            '%s' % (str(scores.mean()), str(scores.std()))
        )
        return scores


def _get_x_y_arrays_from_rois(
        raster_paths, roi_path, spectral_signatures, macroclass=True,
        n_processes: int = None, available_ram: int = None,
        algorithm_name=None, input_normalization=None,
        min_progress=None, max_progress=None
):
    """Gets x y arrays from rois."""
    cfg.logger.log.debug('raster_paths: %s' % str(raster_paths))
    if n_processes is None:
        n_processes = cfg.n_processes
    min_x, max_x, min_y, max_y = raster_vector.get_layer_extent(roi_path)
    virtual_path_list = []
    for p in raster_paths:
        temp_path = cfg.temp.temporary_file_path(name_suffix=cfg.vrt_suffix)
        virtual = raster_vector.create_virtual_raster(
            input_raster_list=[p], output=temp_path,
            box_coordinate_list=[min_x, max_y, max_x, min_y]
        )
        virtual_path_list.append(virtual)
    # get band arrays
    cfg.multiprocess.run_separated(
        raster_path_list=virtual_path_list, function=get_band_arrays,
        function_argument=[[roi_path, spectral_signatures.table]] * len(
            raster_paths
        ), function_variable=virtual_path_list, n_processes=n_processes,
        available_ram=available_ram, keep_output_argument=True,
        progress_message='get band arrays', min_progress=min_progress,
        max_progress=max_progress
    )
    # array for each roi
    cfg.multiprocess.multiprocess_roi_arrays()
    if cfg.multiprocess.output is False:
        return OutputManager(check=False)
    array_dictionary = cfg.multiprocess.output
    x_matrix = y = normalization = means = stds = max_s = min_s = None
    covariance_matrices_dict = {}
    for s in array_dictionary:
        if macroclass:
            class_value = spectral_signatures.table[
                spectral_signatures.table.signature_id == s].macroclass_id[0]
        else:
            class_value = spectral_signatures.table[
                spectral_signatures.table.signature_id == s].class_id[0]
        matrix = np.stack(array_dictionary[s])
        if x_matrix is None:
            x_matrix = matrix
            y = np.ones(matrix.shape[1]) * class_value
        else:
            x_matrix = np.hstack([x_matrix, matrix])
            y = np.concatenate([y, np.ones(matrix.shape[1]) * class_value])
    # normalization
    if input_normalization is not None:
        normalization = []
        means = []
        stds = []
        max_s = []
        min_s = []
        if input_normalization == cfg.z_score:
            for b in range(0, x_matrix.shape[0]):
                mean = x_matrix[b, ::].mean()
                means.append(mean)
                std = x_matrix[b, ::].std()
                stds.append(std)
                expression = '(%s[::, ::, n] - %s) / (%s + 0.000001)' % (
                    cfg.array_function_placeholder, str(mean), str(std))
                x_matrix[b, ::] = (x_matrix[b, ::] - mean) / (std + 0.000001)
                normalization.append(expression)
        elif input_normalization == cfg.linear_scaling:
            for b in range(0, x_matrix.shape[0]):
                minimum = x_matrix[b, ::].min()
                min_s.append(minimum)
                maximum = x_matrix[b, ::].max()
                max_s.append(maximum)
                expression = '(%s[::, ::, n] - %s) / (%s - %s)' % (
                    cfg.array_function_placeholder, str(minimum), str(maximum),
                    str(minimum))
                x_matrix[b, ::] = (x_matrix[b, ::] - minimum) / (
                        maximum - minimum)
                normalization.append(expression)
    if (algorithm_name == cfg.maximum_likelihood
            or algorithm_name == cfg.maximum_likelihood_a):
        for s in array_dictionary:
            cfg.logger.log.error('s: %s' % str(s))
            matrix = np.stack(array_dictionary[s])
            # normalization
            if input_normalization is not None:
                if input_normalization == cfg.z_score:
                    for b in range(0, matrix.shape[0]):
                        matrix[b, ::] = (matrix[b, ::] - means[b]) / (
                                stds[b] + 0.000001)
                elif input_normalization == cfg.linear_scaling:
                    for b in range(0, matrix.shape[0]):
                        matrix[b, ::] = (matrix[b, ::] - min_s[b]) / (
                                max_s[b] - min_s[b])
            # covariance matrix (degree of freedom = 1 for unbiased estimate)
            cov_matrix = np.ma.cov(np.ma.masked_invalid(matrix), ddof=1)
            try:
                # inverse
                inv = np.linalg.inv(cov_matrix)
                if np.isnan(inv[0, 0]):
                    pass
                else:
                    covariance_matrices_dict[s] = cov_matrix
            except Exception as err:
                cfg.logger.log.error(str(err))
                cfg.messages.error('covariance matrix')
    return OutputManager(
        extra={
            'x': x_matrix.T, 'y': y,
            'covariance_matrices': covariance_matrices_dict,
            'normalization_values': [normalization, means, stds, min_s, max_s]
        }
    )


def _collect_x_y_matrices(
        x_matrix=None, y=None, covariance_matrices=None,
        input_raster_list=None, spectral_signatures=None,
        input_normalization=None, macroclass=None, n_processes: int = None,
        available_ram: int = None,
        min_progress=0, max_progress=100, algorithm_name=None
):
    """Collects x and y matrices."""
    normalization_values = None
    if x_matrix is None:
        # calculate x y arrays
        x_y_arrays = _get_x_y_arrays_from_rois(
            raster_paths=input_raster_list,
            roi_path=spectral_signatures.geometry_file,
            spectral_signatures=spectral_signatures, macroclass=macroclass,
            n_processes=n_processes, available_ram=available_ram,
            algorithm_name=algorithm_name,
            input_normalization=input_normalization, min_progress=min_progress,
            max_progress=max_progress
        )
        if x_y_arrays.check is False:
            return OutputManager(check=False)
        x_matrix = x_y_arrays.extra['x']
        y = x_y_arrays.extra['y']
        covariance_matrices = x_y_arrays.extra['covariance_matrices']
        normalization_values = x_y_arrays.extra['normalization_values']
    return OutputManager(
        extra={
            'x': x_matrix, 'y': y, 'covariance_matrices': covariance_matrices,
            'normalization_values': normalization_values
        }
    )


def _create_stratified_k_fold_scikit(classifier, x_matrix, y):
    """Creates stratified k fold for classifier with scikit."""
    classifier_list = []
    train_arg_dict_list = []
    test_arg_dict_list = []
    # stratification
    stratification = StratifiedKFold(n_splits=5, shuffle=True, random_state=0)
    for train_index, test_index in stratification.split(x_matrix, y):
        classifier_list.append(clone(classifier))
        train_arg_dict_list.append(
            {'X': x_matrix[train_index], 'y': y[train_index]}
        )
        test_arg_dict_list.append(
            {'X': x_matrix[test_index], 'y': y[test_index]}
        )
    return classifier_list, train_arg_dict_list, test_arg_dict_list


def _normalize_signatures(
        spectral_signatures, normalization_values, input_normalization
):
    """Normalizes spectral signatures."""
    if input_normalization is not None:
        x_signatures = deepcopy(spectral_signatures.signatures)
        for s in x_signatures:
            if input_normalization == cfg.z_score:
                x_signatures[s].value = ((x_signatures[s].value - np.array(
                    normalization_values[1]
                )) / (np.array(
                    normalization_values[2]
                ) + 0.000001))
            elif input_normalization == cfg.linear_scaling:
                x_signatures[s].value = (x_signatures[s].value - np.array(
                    normalization_values[3]
                )) / (np.array(
                    normalization_values[4]
                ) - np.array(
                    normalization_values[3]
                ))
        spectral_signatures_catalog = deepcopy(spectral_signatures)
        spectral_signatures_catalog.signatures = x_signatures
    else:
        spectral_signatures_catalog = spectral_signatures
    return spectral_signatures_catalog


def _load_model(model_path):
    """Loads a model."""
    cfg.logger.log.debug('model_path: %s' % str(model_path))
    temp_dir = cfg.temp.create_temporary_directory()
    file_list = files_directories.unzip_file(model_path, temp_dir)
    algorithm_name = framework_name = spectral_signatures = \
        covariance_matrices = model_classifier = None
    input_normalization = normalization_values = None
    # open classification framework
    for f in file_list:
        f_name = files_directories.file_name(f)
        if f_name == cfg.classification_framework:
            classification_framework = read_write_files.open_text_file(f)
            lines = classification_framework.split(cfg.new_line)
            for line in lines:
                variable = line.split('=')
                if variable[0] == cfg.algorithm_name_framework:
                    algorithm_name = variable[1]
                elif variable[0] == cfg.input_normalization_framework:
                    input_normalization = variable[1]
            if input_normalization is not None and \
                    input_normalization.lower() == 'none':
                input_normalization = None
            # scikit framework
            if (
                    algorithm_name == cfg.random_forest
                    or algorithm_name == cfg.random_forest_a
                    or algorithm_name == cfg.random_forest_ovr
                    or algorithm_name == cfg.random_forest_ovr_a
                    or algorithm_name == cfg.support_vector_machine
                    or algorithm_name == cfg.support_vector_machine_a
                    or algorithm_name == cfg.multi_layer_perceptron
                    or algorithm_name == cfg.multi_layer_perceptron_a):
                framework_name = cfg.scikit_framework
            # pytorch framework
            elif (algorithm_name == cfg.pytorch_multi_layer_perceptron
                  or algorithm_name == cfg.pytorch_multi_layer_perceptron_a):
                framework_name = cfg.pytorch_framework
        elif f_name == cfg.spectral_signatures_framework:
            spectral_signatures = pickle.load(open(f, 'rb'))
        elif f_name == cfg.normalization_values_framework:
            normalization_values = pickle.load(open(f, 'rb'))
        elif f_name == cfg.covariance_matrices_framework:
            covariance_matrices = pickle.load(open(f, 'rb'))
    # open model files
    for f in file_list:
        f_name = files_directories.file_name(f)
        if f_name == cfg.model_classifier_framework:
            if framework_name == cfg.scikit_framework:
                model_classifier = pickle.load(open(f, 'rb'))
            elif framework_name == cfg.pytorch_framework:
                model_classifier = torch.load(f)
    classifier = Classifier.load_classifier(
        algorithm_name=algorithm_name, spectral_signatures=spectral_signatures,
        covariance_matrices=covariance_matrices,
        model_classifier=model_classifier,
        input_normalization=input_normalization,
        normalization_values=normalization_values
    )
    return OutputManager(extra={'classifier': classifier})



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

"""
Tools to manage operating systems
"""

from platform import system
import sys

from remotior_sensus.core import configurations as cfg


# get system information
def get_system_info():
    if sys.maxsize > 2 ** 32:
        cfg.sys_64bit = True
    else:
        cfg.sys_64bit = False
    # file system encoding
    cfg.file_sys_encoding = sys.getfilesystemencoding()
    # system information
    cfg.sys_name = system()
    cfg.logger.log.info(
        'system: %s; 64bit: %s; n_processes: %s; ram: %s; temp.dir: %s'
        % (cfg.sys_name, cfg.sys_64bit, cfg.n_processes, cfg.available_ram,
           cfg.temp.dir)
        )



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

"""
Perform the preprocessing of products.
"""

import os
from xml.dom import minidom
from typing import Optional

import numpy as np

from remotior_sensus.core import configurations as cfg, table_manager as tm
from remotior_sensus.core.bandset_catalog import BandSetCatalog
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.core.processor_functions import (
    band_calculation, raster_unique_values_with_sum
)
from remotior_sensus.util import files_directories


# create product table and preprocess
def preprocess(
        input_path, output_path, metadata_file_path=None, add_bandset=None,
        product=None, nodata_value=None, sensor=None, acquisition_date=None,
        dos1_correction=False, output_prefix='', n_processes: int = None,
        bandset_catalog: Optional[BandSetCatalog] = None,
        available_ram: int = None, progress_message=True
) -> OutputManager:
    """Create table and preprocess products.

    Perform image conversion to reflectance of several products.

    Can calculate DOS1 corrected reflectance (Sobrino, J. et al., 2004. Land
    surface temperature retrieval from LANDSAT TM 5. Remote Sensing of
    Environment, Elsevier, 90, 434-440)  approximating path radiance
    to path reflectance for level 1 data:
    TOA reflectance = DN * reflectance_scale + reflectance_offset
    path reflectance p = DNm - Dark Object reflectance = DNm * reflectance_scale + reflectance_offset - 0.01
    land surface reflectance = TOA reflectance - p = (DN * reflectance_scale) - (DNm * reflectance_scale - 0.01)

    Landsat's data Collection 1 and 2
    Level 1T
    Landsat 8-9 TOA reflectance proportional to exo-atmospheric solar
    irradiance in each band and the Earth-Sun distance
    (USGS, 2021. Landsat 8-9 Calibration and Validation (Cal/Val) Algorithm
    Description Document (ADD). Version 4.0. Department of the Interior
    U.S. Geological Survey, South Dakota)
    TOA reflectance with correction for the sun angle =
    DN * Reflectance multiplicative scaling factor + Reflectance additive
    scaling factor / sin(Sun elevation)
    Level 2S
    Surface reflectance = DN * Reflectance multiplicative scaling factor +
    Reflectance additive scaling factor

    Sentinel-2 data
    Level 1C
    TOA reflectance = DN / QUANTIFICATION VALUE + OFFSET
    Level 2S
    Surface reflectance = DN / QUANTIFICATION VALUE + OFFSET

    Args:
        input_path:
        output_path: string of output path directory.
        metadata_file_path:
        dos1_correction: if True, perform DOS1 correction.
        add_bandset: if True, create a new bandset and add output bands to it; if False, add output bands to current bandset.
        product:
        nodata_value:
        sensor:
        acquisition_date:
        output_prefix: optional string for output name prefix.
        n_processes: number of parallel processes.
        available_ram: number of megabytes of RAM available to processes.
        bandset_catalog: BandSetCatalog object.
        progress_message: progress message.

    Returns:
        Object :func:`~remotior_sensus.core.output_manager.OutputManager` with
            - paths = output list
    """  # noqa: E501
    table = create_product_table(
        input_path=input_path, metadata_file_path=metadata_file_path,
        product=product, nodata_value=nodata_value, sensor=sensor,
        acquisition_date=acquisition_date
    )
    output = perform_preprocess(
        product_table=table, output_path=output_path, add_bandset=add_bandset,
        dos1_correction=dos1_correction, output_prefix=output_prefix,
        n_processes=n_processes, available_ram=available_ram,
        progress_message=progress_message, bandset_catalog=bandset_catalog
        )
    return output


# preprocess products
def perform_preprocess(
        product_table, output_path, dos1_correction=False, add_bandset=None,
        output_prefix='', n_processes: int = None, available_ram: int = None,
        bandset_catalog: Optional[BandSetCatalog] = None, progress_message=True
) -> OutputManager:
    """Preprocess products.

    Perform preprocessing based on product table.

    Args:
        product_table: product table object.
        output_path: string of output path directory.
        dos1_correction: if True, perform DOS1 correction.
        add_bandset: if True, create a new bandset and add output bands to it; if False, add output bands to current bandset.
        output_prefix: optional string for output name prefix.
        n_processes: number of parallel processes.
        available_ram: number of megabytes of RAM available to processes.
        bandset_catalog: BandSetCatalog object.
        progress_message: progress message.

    Returns:
        Object :func:`~remotior_sensus.core.output_manager.OutputManager` with
            - paths = output list
    """  # noqa: E501
    if progress_message:
        cfg.logger.log.info('start')
        cfg.progress.update(
            process=__name__.split('.')[-1].replace('_', ' '),
            message='starting', start=True
        )
    cfg.logger.log.debug('product_table: %s' % str(product_table))
    if n_processes is None:
        n_processes = cfg.n_processes
    input_list = []
    input_dos1_list = []
    dos1_nodata_list = []
    nodata_list = []
    calculation_datatype = []
    scale_list = []
    offset_list = []
    output_nodata = []
    output_datatype = []
    output_raster_path_list = []
    expressions = []
    dos1_expressions = []
    # create process string list
    # Sentinel-2
    sentinel_product = product_table[product_table.product == cfg.sentinel2]
    # Landsat
    landsat_product = product_table[product_table.product == cfg.landsat]
    # Sentinel-2
    if len(sentinel_product) > 0:
        if dos1_correction:
            # exclude Level-2A
            sentinel_product_1c = sentinel_product[
                np.char.lower(sentinel_product.processing_level) != 'level-2a']
            # calculate DOS1 corrected reflectance approximating path
            # radiance to path reflectance
            # land surface reflectance = TOA reflectance - p =
            # (DN * reflectance_scale) - (DNm * reflectance_scale - 0.01)
            # raster and dnm are variables in the calculation
            string_1 = np.char.add(
                'np.clip(( %s * ' % cfg.array_function_placeholder,
                sentinel_product_1c.scale.astype('<U16')
            )
            string_2 = np.char.add(string_1, ' - (')
            string_3 = np.char.add(
                string_2, sentinel_product_1c.scale.astype('<U16')
            )
            dos1_expressions.extend(
                np.char.add(string_3, ' * dnm - 0.01)), 0, 1)').tolist()
            )
            input_dos1_list.extend(sentinel_product_1c.product_path.tolist())
            # output raster list
            output_string_1 = np.char.add(
                '%s/%s' % (output_path, output_prefix),
                sentinel_product_1c.band_name
            )
            output_raster_path_list.extend(
                np.char.add(output_string_1, cfg.tif_suffix).tolist()
            )
            nodata_list.extend(sentinel_product_1c.nodata.tolist())
            dos1_nodata_list.extend(sentinel_product_1c.nodata.tolist())
            calculation_datatype.extend(
                [np.float32] * len(sentinel_product_1c)
            )
            output_datatype.extend([cfg.uint16_dt] * len(sentinel_product_1c))
            scale_list.extend([0.0001] * len(sentinel_product_1c))
            offset_list.extend([0] * len(sentinel_product_1c))
            output_nodata.extend(
                [cfg.nodata_val_UInt16] * len(sentinel_product_1c)
            )
            # get Level-2A if any and perform conversion without DOS1
            sentinel_product_2a = sentinel_product[
                np.char.lower(sentinel_product.processing_level) == 'level-2a']
            # calculate reflectance = DN / quantificationValue = DN * scale
            # raster is interpreted as variable in the calculation
            string_1 = np.char.add(
                'np.clip( ( %s * ' % cfg.array_function_placeholder,
                sentinel_product_2a.scale.astype('<U16')
            )
            expressions.extend(np.char.add(string_1, ') , 0, 1)').tolist())
            input_list.extend(sentinel_product_2a.product_path.tolist())
            # output raster list
            output_string_1 = np.char.add(
                '%s/%s' % (output_path, output_prefix),
                sentinel_product_2a.band_name
            )
            output_raster_path_list.extend(
                np.char.add(output_string_1, cfg.tif_suffix).tolist()
            )
            nodata_list.extend(sentinel_product_2a.nodata.tolist())
            calculation_datatype.extend(
                [np.float32] * len(sentinel_product_2a)
            )
            output_datatype.extend([cfg.uint16_dt] * len(sentinel_product_2a))
            scale_list.extend([0.0001] * len(sentinel_product_2a))
            offset_list.extend([0] * len(sentinel_product_2a))
            output_nodata.extend(
                [cfg.nodata_val_UInt16] * len(sentinel_product_2a)
            )
        else:
            # calculate reflectance = DN / quantificationValue = DN * scale
            # raster is interpreted as variable in the calculation
            string_1 = np.char.add(
                'np.clip( ( %s * ' % cfg.array_function_placeholder,
                sentinel_product.scale.astype('<U16')
            )
            expressions.extend(np.char.add(string_1, ') , 0, 1)').tolist())
            input_list.extend(sentinel_product.product_path.tolist())
            # output raster list
            output_string_1 = np.char.add(
                '%s/%s' % (output_path, output_prefix),
                sentinel_product.band_name
            )
            output_raster_path_list.extend(
                np.char.add(output_string_1, cfg.tif_suffix).tolist()
            )
            nodata_list.extend(sentinel_product.nodata.tolist())
            calculation_datatype.extend([np.float32] * len(sentinel_product))
            output_datatype.extend([cfg.uint16_dt] * len(sentinel_product))
            scale_list.extend([0.0001] * len(sentinel_product))
            offset_list.extend([0] * len(sentinel_product))
            output_nodata.extend(
                [cfg.nodata_val_UInt16] * len(sentinel_product)
            )
    # Landsat
    elif len(landsat_product) > 0:
        # temperature
        landsat_temperature_product = landsat_product[landsat_product.k1 != 0]
        string_0 = np.char.add(
            landsat_temperature_product.k2.astype('<U16'), ' / ( log( 1 + '
        )
        string_1 = np.char.add(
            string_0, landsat_temperature_product.k1.astype('<U16')
        )
        string_2 = np.char.add(
            string_1, ' / (%s * ' % cfg.array_function_placeholder
        )
        string_3 = np.char.add(
            string_2, landsat_temperature_product.scale.astype('<U16')
        )
        string_4 = np.char.add(string_3, ' + ')
        string_5 = np.char.add(
            string_4, landsat_temperature_product.offset.astype('<U16')
        )
        expressions.extend(np.char.add(string_5, ') ) )').tolist())
        input_list.extend(landsat_temperature_product.product_path.tolist())
        calculation_datatype.extend(
            [np.float32] * len(landsat_temperature_product)
        )
        output_datatype.extend(
            [cfg.float32_dt] * len(landsat_temperature_product)
        )
        scale_list.extend([1] * len(landsat_temperature_product))
        offset_list.extend([0] * len(landsat_temperature_product))
        output_nodata.extend(
            [cfg.nodata_val_Float32] * len(landsat_temperature_product)
        )
        # output raster list
        output_string_temperature_1 = np.char.add(
            '%s/%s' % (output_path, output_prefix),
            landsat_temperature_product.band_name
        )
        output_raster_path_list.extend(
            np.char.add(output_string_temperature_1, cfg.tif_suffix).tolist()
        )
        nodata_list.extend(landsat_temperature_product.nodata.tolist())
        landsat_temperature_product_10 = landsat_product[
            landsat_product.band_number == '10']
        string_1 = np.char.add(
            '%s * ' % cfg.array_function_placeholder,
            landsat_temperature_product_10.scale.astype('<U16')
        )
        string_2 = np.char.add(string_1, ' + ')
        string_3 = np.char.add(
            string_2, landsat_temperature_product_10.offset.astype('<U16')
        )
        expressions.extend(string_3.tolist())
        input_list.extend(landsat_temperature_product_10.product_path.tolist())
        calculation_datatype.extend(
            [np.float32] * len(landsat_temperature_product_10)
        )
        output_datatype.extend(
            [cfg.float32_dt] * len(landsat_temperature_product_10)
        )
        scale_list.extend([1] * len(landsat_temperature_product_10))
        offset_list.extend([0] * len(landsat_temperature_product_10))
        output_nodata.extend(
            [cfg.nodata_val_Float32] * len(landsat_temperature_product_10)
        )
        # output raster list
        output_string_temperature_10 = np.char.add(
            '%s/%s' % (output_path, output_prefix),
            landsat_temperature_product_10.band_name
        )
        output_raster_path_list.extend(
            np.char.add(output_string_temperature_10, cfg.tif_suffix).tolist()
        )
        nodata_list.extend(landsat_temperature_product_10.nodata.tolist())
        if dos1_correction:
            # exclude level 2 products and temperature
            landsat_product_l1 = landsat_product[
                (np.char.lower(landsat_product.processing_level) == 'l1tp') & (
                        landsat_product.band_number != '10') &
                (landsat_product.k1 == 0)]
            # calculate DOS1 corrected reflectance approximating path
            # radiance to path reflectance
            # land surface reflectance = TOA reflectance - p =
            # (DN * reflectance_scale) - (DNm * reflectance_scale - 0.01)
            # raster and dnm are variables in the calculation
            string_1 = np.char.add(
                'np.clip(( %s * ' % cfg.array_function_placeholder,
                landsat_product_l1.scale.astype('<U16')
            )
            string_2 = np.char.add(string_1, ' - (')
            string_3 = np.char.add(
                string_2, landsat_product_l1.scale.astype('<U16')
            )
            dos1_expressions.extend(
                np.char.add(string_3, ' * dnm - 0.01)), 0, 1)').tolist()
            )
            input_dos1_list.extend(landsat_product_l1.product_path.tolist())
            # output raster list
            output_string_1 = np.char.add(
                '%s/%s' % (output_path, output_prefix),
                landsat_product_l1.band_name
            )
            output_raster_path_list.extend(
                np.char.add(output_string_1, cfg.tif_suffix).tolist()
            )
            nodata_list.extend(landsat_product_l1.nodata.tolist())
            dos1_nodata_list.extend(landsat_product_l1.nodata.tolist())
            calculation_datatype.extend([np.float32] * len(landsat_product_l1))
            output_datatype.extend([cfg.uint16_dt] * len(landsat_product_l1))
            scale_list.extend([0.0001] * len(landsat_product_l1))
            offset_list.extend([0] * len(landsat_product_l1))
            output_nodata.extend(
                [cfg.nodata_val_UInt16] * len(landsat_product_l1)
            )
        else:
            # level 1 products
            landsat_1_product = landsat_product[
                (np.char.lower(landsat_product.processing_level) == 'l1tp') & (
                        landsat_product.band_number != '10')
                & (landsat_product.k1 == 0)]
            # calculate reflectance = (raster * scale
            #  + offset) / sin(Sun elevation)
            # raster is interpreted as variable in the calculation
            string_1 = np.char.add(
                'np.clip( ( ( %s * ' % cfg.array_function_placeholder,
                landsat_1_product.scale.astype('<U16')
            )
            string_2 = np.char.add(string_1, ' + ')
            string_3 = np.char.add(
                string_2, landsat_1_product.offset.astype('<U16')
            )
            string_4 = np.char.add(string_3, ') / sin(')
            string_5 = np.char.add(
                string_4, landsat_1_product.sun_elevation.astype(
                    '<U16'
                )
            )
            expressions.extend(np.char.add(string_5, ') ) , 0, 1)').tolist())
            input_list.extend(landsat_1_product.product_path.tolist())
            # output raster list
            output_string_1 = np.char.add(
                '%s/%s' % (output_path, output_prefix),
                landsat_1_product.band_name
            )
            output_raster_path_list.extend(
                np.char.add(output_string_1, cfg.tif_suffix).tolist()
            )
            nodata_list.extend(landsat_1_product.nodata.tolist())
            calculation_datatype.extend([np.float32] * len(landsat_1_product))
            output_datatype.extend([cfg.uint16_dt] * len(landsat_1_product))
            scale_list.extend([0.0001] * len(landsat_1_product))
            offset_list.extend([0] * len(landsat_1_product))
            output_nodata.extend(
                [cfg.nodata_val_UInt16] * len(landsat_1_product)
            )
            # level 2 products
            landsat_2_product = landsat_product[
                (np.char.lower(landsat_product.processing_level) == 'l2sp') & (
                        landsat_product.band_number != '10')]
            # calculate reflectance = (raster * scale
            #  + offset) / sin(Sun elevation)
            # raster is interpreted as variable in the calculation
            string_1 = np.char.add(
                'np.clip( ( %s * ' % cfg.array_function_placeholder,
                landsat_2_product.scale.astype('<U16')
            )
            string_2 = np.char.add(string_1, ' + ')
            string_3 = np.char.add(
                string_2, landsat_2_product.offset.astype('<U16')
            )
            expressions.extend(np.char.add(string_3, ') , 0, 1)').tolist())
            input_list.extend(landsat_2_product.product_path.tolist())
            # output raster list
            output_string_2 = np.char.add(
                '%s/%s' % (output_path, output_prefix),
                landsat_2_product.band_name
            )
            output_raster_path_list.extend(
                np.char.add(output_string_2, cfg.tif_suffix).tolist()
            )
            nodata_list.extend(landsat_2_product.nodata.tolist())
            calculation_datatype.extend([np.float32] * len(landsat_product))
            output_datatype.extend([cfg.uint16_dt] * len(landsat_product))
            scale_list.extend([0.0001] * len(landsat_product))
            offset_list.extend([0] * len(landsat_product))
            output_nodata.extend(
                [cfg.nodata_val_UInt16] * len(landsat_product)
            )
    files_directories.create_directory(output_path)
    # dummy bands for memory calculation
    dummy_bands = 2
    # conversion
    if dos1_correction:
        # get min dn values
        cfg.multiprocess.run_separated(
            raster_path_list=input_dos1_list,
            function=raster_unique_values_with_sum, dummy_bands=dummy_bands,
            use_value_as_nodata=dos1_nodata_list, n_processes=n_processes,
            available_ram=available_ram, keep_output_argument=True,
            progress_message='unique values', min_progress=1, max_progress=30
        )
        cfg.multiprocess.find_minimum_dn()
        if cfg.multiprocess.output is False:
            cfg.logger.log.error('unable to calculate')
            cfg.messages.error('unable to calculate')
            return OutputManager(check=False)
        min_dn = cfg.multiprocess.output
        for i in range(len(dos1_expressions)):
            expressions.append(
                dos1_expressions[i].replace('dnm', str(min_dn[i]))
            )
            input_list.append(input_dos1_list[i])
    # dummy bands for memory calculation
    dummy_bands = 2
    # run calculation
    cfg.multiprocess.run_separated(
        raster_path_list=input_list, function=band_calculation,
        function_argument=expressions,
        calculation_datatype=calculation_datatype,
        use_value_as_nodata=nodata_list, dummy_bands=dummy_bands,
        output_raster_list=output_raster_path_list,
        output_data_type=output_datatype, output_nodata_value=output_nodata,
        compress=cfg.raster_compression, n_processes=n_processes,
        available_ram=available_ram, scale=scale_list, offset=offset_list,
        progress_message='processing', min_progress=30, max_progress=99
    )
    if len(output_raster_path_list) == 0:
        cfg.logger.log.error('unable to process files')
        cfg.messages.error('unable to process files')
        return OutputManager(check=False)
    else:
        for i in output_raster_path_list:
            if not files_directories.is_file(i):
                cfg.logger.log.error('unable to process file: %s' % str(i))
                cfg.messages.error('unable to process file: %s' % str(i))
                return OutputManager(check=False)
    # add output to BandSet
    if add_bandset is not None and bandset_catalog is not None:
        product = product_table.product[0]
        # get Landsat satellite
        if product == cfg.landsat:
            spacecraft = product_table.spacecraft[0]
            if '1' in spacecraft or '2' in spacecraft or '3' in spacecraft:
                product = cfg.satLandsat13
            elif '4' in spacecraft or '5' in spacecraft:
                product = cfg.satLandsat45
            elif '7' in spacecraft:
                product = cfg.satLandsat7
            elif '8' in spacecraft:
                product = cfg.satLandsat8
            elif '9' in spacecraft:
                product = cfg.satLandsat9
        if add_bandset is True:
            bandset_number = bandset_catalog.get_bandset_count() + 1
            # create bandset
            bandset_catalog.create_bandset(
                paths=output_raster_path_list, wavelengths=[product],
                insert=True, date=str(product_table.date[0]),
                bandset_number=bandset_number
            )
            bandset_catalog.get_bandset(
                bandset_number).sort_bands_by_wavelength()
        else:
            for path in output_raster_path_list:
                # add to current bandset
                try:
                    bandset_catalog.add_band_to_bandset(
                        path=path,
                        bandset_number=bandset_catalog.current_bandset
                    )
                    bandset_catalog.set_satellite_wavelength(
                        satellite_name=product,
                        bandset_number=bandset_catalog.current_bandset)
                except Exception as err:
                    cfg.logger.log.error(str(err))
                    cfg.messages.error(str(err))
    cfg.progress.update(end=True)
    cfg.logger.log.info(
        'end; preprocess products: %s' % str(output_raster_path_list)
    )
    return OutputManager(paths=output_raster_path_list)


# create product table
def create_product_table(
        input_path, metadata_file_path=None, product=None, nodata_value=None,
        sensor=None, acquisition_date=None
):
    band_names = []
    band_number_list = []
    product_path_list = []
    product_name_list = []
    scale_value_list = []
    scale_offset_dict = {}
    k_dict = {}
    offset_value_list = []
    k2_list = []
    k1_list = []
    e_sun_list = []
    sun_elevation_list = []
    earth_sun_distance_list = []
    spacecraft_list = []
    metadata = metadata_doc = metadata_type = product_date = product_name = \
        processing_level = sun_elevation = None
    earth_sun_distance = None
    if product == cfg.sentinel2:
        product_name = cfg.sentinel2
    elif product == cfg.landsat:
        product_name = cfg.landsat
    # get metadata
    if metadata_file_path is None:
        for f in os.listdir(input_path):
            # Sentinel-2 metadata
            if f.lower().endswith('.xml') and (
                    'mtd_msil1c' in f.lower() or 'mtd_safl1c' in f.lower()
                    or 'mtd_msil2a' in f.lower()):
                metadata = '%s/%s' % (input_path, str(f))
                product_name = cfg.sentinel2
                metadata_type = 'xml'
            # Landsat metadata
            elif f[0].lower() == 'l' and f.lower().endswith(
                    '.xml'
            ) and 'mtl' in f.lower():
                metadata = '%s/%s' % (input_path, str(f))
                product_name = cfg.landsat
                metadata_type = 'xml'
    else:
        metadata = metadata_file_path
        if files_directories.file_extension(metadata) == cfg.xml_suffix:
            metadata_type = 'xml'
    if metadata_type == 'xml':
        # open metadata
        try:
            metadata_doc = minidom.parse(metadata)
            # Sentinel-2
            try:
                spacecraft_name = \
                    metadata_doc.getElementsByTagName('SPACECRAFT_NAME')[
                        0].firstChild.data
                if spacecraft_name:
                    product_name = cfg.sentinel2
            except Exception as err:
                str(err)
            # Landsat
            try:
                spacecraft_id = \
                    metadata_doc.getElementsByTagName('SPACECRAFT_ID')[
                        0].firstChild.data
                if spacecraft_id:
                    product_name = cfg.landsat
            except Exception as err:
                str(err)
        except Exception as err:
            cfg.messages.error('unable to open metadata')
            cfg.logger.log.error(str(err))
            return OutputManager(check=False)
    # Sentinel-2
    if product_name == cfg.sentinel2:
        cfg.logger.log.debug(cfg.sentinel2)
        scale_value = 1 / 10000
        offset_value = 0
        sentinel2_bands = cfg.satellites[cfg.satSentinel2][2]
        # open metadata
        if metadata_doc:
            try:
                # get date in the format YYYY-MM-DD
                product_date = (
                    metadata_doc.getElementsByTagName(
                        'PRODUCT_START_TIME')[0].firstChild.data.split('T')[0]
                )
                processing_level = (
                    metadata_doc.getElementsByTagName(
                        'PROCESSING_LEVEL')[0].firstChild.data
                )
                # L2A products
                if '2a' in processing_level.lower():
                    scale_value = 1 / int(
                        metadata_doc.getElementsByTagName(
                            'BOA_QUANTIFICATION_VALUE'
                        )[0].firstChild.data
                    )
                    offset = metadata_doc.getElementsByTagName(
                        'BOA_ADD_OFFSET'
                    )
                # L1C products
                else:
                    scale_value = 1 / int(
                        metadata_doc.getElementsByTagName(
                            'QUANTIFICATION_VALUE'
                        )[0].firstChild.data
                    )
                    offset = metadata_doc.getElementsByTagName(
                        'RADIO_ADD_OFFSET'
                    )
                for n in range(len(sentinel2_bands)):
                    if offset_value:
                        offset_value_list.append(offset[n].firstChild.data)
                cfg.logger.log.debug('metadata')
            except Exception as err:
                cfg.logger.log.error(str(err))
                cfg.messages.error(str(err))
        # use default values
        else:
            scale_value = 1 / 10000
            offset_value_list = [0] * len(sentinel2_bands)
            cfg.messages.warning('using default values without metadata')
            cfg.logger.log.debug('no metadata')
        # get bands
        file_list = files_directories.files_in_directory(
            input_path, sort_files=True, suffix_filter=cfg.tif_suffix
        )
        file_list.extend(
            files_directories.files_in_directory(
                input_path, sort_files=True, suffix_filter='.jp2'
            )
        )
        file_list.extend(
            files_directories.files_in_directory(
                input_path, sort_files=True, suffix_filter='.vrt'
            )
        )
        for f in file_list:
            # check band number
            if f[-6:-4].lower() in sentinel2_bands:
                band_names.append(files_directories.file_name(f))
                product_path_list.append(f)
                band_number_list.append(f[-6:-4].lower())
        product_name_list = [product_name] * len(band_names)
        spacecraft_list = [product_name] * len(band_names)
        scale_value_list = [scale_value] * len(band_names)
        if len(offset_value_list) == 0:
            offset_value_list = [0] * len(band_names)
    elif product_name == cfg.landsat:
        cfg.logger.log.debug(cfg.landsat)
        # open metadata
        if metadata_doc:
            sensor_id = metadata_doc.getElementsByTagName('SENSOR_ID')[
                0].firstChild.data
            processing_level = (
                metadata_doc.getElementsByTagName(
                    'PROCESSING_LEVEL')[0].firstChild.data.lower()
            )
            spacecraft_id = metadata_doc.getElementsByTagName('SPACECRAFT_ID')[
                0].firstChild.data.lower()
            # get date in the format YYYY-MM-DD
            product_date = metadata_doc.getElementsByTagName('DATE_ACQUIRED')[
                0].firstChild.data
            sun_elevation = metadata_doc.getElementsByTagName('SUN_ELEVATION')[
                0].firstChild.data
            earth_sun_distance = (
                metadata_doc.getElementsByTagName(
                    'EARTH_SUN_DISTANCE')[0].firstChild.data
            )
            if sensor_id.lower() == cfg.sensor_oli:
                band_list = cfg.satellites[cfg.satLandsat8][2]
            elif sensor_id.lower() == cfg.sensor_etm:
                band_list = cfg.satellites[cfg.satLandsat7][2]
            elif sensor_id.lower() == cfg.sensor_tm:
                band_list = cfg.satellites[cfg.satLandsat45][2]
            elif sensor_id.lower() == cfg.sensor_mss:
                band_list = cfg.satellites[cfg.satLandsat13][2]
            else:
                band_list = []
            for b in band_list:
                if processing_level == 'l2sp':
                    reflectance_tag = metadata_doc.getElementsByTagName(
                        'LEVEL2_SURFACE_REFLECTANCE_PARAMETERS'
                    )[0]
                else:
                    reflectance_tag = metadata_doc.getElementsByTagName(
                        'LEVEL1_RADIOMETRIC_RESCALING'
                    )[0]
                try:
                    reflectance_mult = reflectance_tag.getElementsByTagName(
                        'REFLECTANCE_MULT_BAND_%s' % str(b)
                    )[0].firstChild.data
                    reflectance_add = reflectance_tag.getElementsByTagName(
                        'REFLECTANCE_ADD_BAND_%s' % str(b)
                    )[0].firstChild.data
                    scale_offset_dict[str(b)] = [float(reflectance_mult),
                                                 float(reflectance_add)]
                except Exception as err:
                    str(err)
                    scale_offset_dict[str(b)] = [1, 0]
            # temperature
            # Landsat 8-9
            if (sensor_id.lower() == cfg.sensor_oli
                    and processing_level == 'l2sp'):
                scale_offset_dict['10'] = [float(
                    metadata_doc.getElementsByTagName(
                        'TEMPERATURE_MULT_BAND_ST_B10'
                    )[0].firstChild.data
                ), float(
                    metadata_doc.getElementsByTagName(
                        'TEMPERATURE_ADD_BAND_ST_B10'
                    )[0].firstChild.data
                )]
            # Landsat 7 level 1
            elif (sensor_id.lower() == cfg.sensor_etm
                  and processing_level == 'l1tp'):
                k_dict['6_VCID_1'] = [float(
                    metadata_doc.getElementsByTagName(
                        'K1_CONSTANT_BAND_6_VCID_1'
                    )[0].firstChild.data
                ), float(
                    metadata_doc.getElementsByTagName(
                        'K2_CONSTANT_BAND_6_VCID_1'
                    )[0].firstChild.data
                ), float(
                    metadata_doc.getElementsByTagName(
                        'RADIANCE_MULT_BAND_6_VCID_1'
                    )[0].firstChild.data
                ), float(
                    metadata_doc.getElementsByTagName(
                        'RADIANCE_ADD_BAND_6_VCID_1'
                    )[0].firstChild.data
                )]
                k_dict['6_VCID_2'] = [float(
                    metadata_doc.getElementsByTagName(
                        'K1_CONSTANT_BAND_6_VCID_2'
                    )[0].firstChild.data
                ), float(
                    metadata_doc.getElementsByTagName(
                        'K2_CONSTANT_BAND_6_VCID_2'
                    )[0].firstChild.data
                ), float(
                    metadata_doc.getElementsByTagName(
                        'RADIANCE_MULT_BAND_6_VCID_2'
                    )[0].firstChild.data
                ), float(
                    metadata_doc.getElementsByTagName(
                        'RADIANCE_ADD_BAND_6_VCID_2'
                    )[0].firstChild.data
                )]
            # Landsat 7 level 2
            elif (sensor_id.lower() == cfg.sensor_tm
                  and processing_level == 'l2sp'):
                scale_offset_dict['6'] = [float(
                    metadata_doc.getElementsByTagName(
                        'TEMPERATURE_MULT_BAND_ST_B6'
                    )[0].firstChild.data
                ), float(
                    metadata_doc.getElementsByTagName(
                        'TEMPERATURE_ADD_BAND_ST_B6'
                    )[0].firstChild.data
                )]
            # Landsat 5 level 1
            elif (sensor_id.lower() == cfg.sensor_tm
                  and processing_level == 'l1tp'):
                k_dict['6'] = [float(
                    metadata_doc.getElementsByTagName('K1_CONSTANT_BAND_6')[
                        0].firstChild.data
                ), float(
                    metadata_doc.getElementsByTagName('K2_CONSTANT_BAND_6')[
                        0].firstChild.data
                ), float(
                    metadata_doc.getElementsByTagName('RADIANCE_MULT_BAND_6')[
                        0].firstChild.data
                ), float(
                    metadata_doc.getElementsByTagName('RADIANCE_ADD_BAND_6')[
                        0].firstChild.data
                )]
            # Landsat 5 level 2
            elif (sensor_id.lower() == cfg.sensor_tm
                  and processing_level == 'l2sp'):
                scale_offset_dict['6'] = [float(
                    metadata_doc.getElementsByTagName(
                        'TEMPERATURE_MULT_BAND_ST_B6'
                    )[0].firstChild.data
                ), float(
                    metadata_doc.getElementsByTagName(
                        'TEMPERATURE_ADD_BAND_ST_B6'
                    )[0].firstChild.data
                )]

        # use default values
        else:
            spacecraft_id = cfg.landsat
            if sensor:
                sensor_id = sensor
            else:
                sensor_id = 'oli_tirs'
                processing_level = 'l2sp'
            if acquisition_date:
                product_date = acquisition_date
            else:
                product_date = '2000-01-01'
            cfg.messages.warning('using default values without metadata')
            cfg.logger.log.debug('no metadata')
        cfg.logger.log.debug('sensor_id: %s' % sensor_id)
        # get bands
        if sensor_id.lower() == cfg.sensor_oli:
            landsat_bands = cfg.satellites[cfg.satLandsat8][2]
        elif sensor_id.lower() == cfg.sensor_etm:
            landsat_bands = cfg.satellites[cfg.satLandsat7][2]
        elif sensor_id.lower() == cfg.sensor_tm:
            landsat_bands = cfg.satellites[cfg.satLandsat45][2]
        elif sensor_id.lower() == cfg.sensor_mss:
            landsat_bands = cfg.satellites[cfg.satLandsat13][2]
        else:
            landsat_bands = []
        file_list = files_directories.files_in_directory(
            input_path, sort_files=True, suffix_filter=cfg.tif_suffix
        )
        for f in file_list:
            # check band number for multispectral bands
            if f[-5:-4] in landsat_bands:
                band_names.append(files_directories.file_name(f))
                band_number_list.append(f[-5:-4])
                product_path_list.append(f)
                sun_elevation_list.append(sun_elevation)
                earth_sun_distance_list.append(earth_sun_distance)
                if f[-5:-4] in scale_offset_dict:
                    scale_value_list.append(scale_offset_dict[f[-5:-4]][0])
                    offset_value_list.append(scale_offset_dict[f[-5:-4]][1])
                k1_list.append(0)
                k2_list.append(0)
            # temperature bands Landsat 5 band 6
            elif sensor_id.lower() == cfg.sensor_tm:
                if processing_level == 'l2sp' and f[-5:-4] == '6':
                    band_names.append(files_directories.file_name(f))
                    band_number_list.append('6')
                    product_path_list.append(f)
                    sun_elevation_list.append(sun_elevation)
                    earth_sun_distance_list.append(earth_sun_distance)
                    if '6' in scale_offset_dict:
                        scale_value_list.append(scale_offset_dict['6'][0])
                        offset_value_list.append(scale_offset_dict['6'][1])
                    k1_list.append(0)
                    k2_list.append(0)
                elif processing_level == 'l1tp' and f[-5:-4] == '6':
                    band_names.append(files_directories.file_name(f))
                    band_number_list.append('6')
                    product_path_list.append(f)
                    sun_elevation_list.append(sun_elevation)
                    earth_sun_distance_list.append(earth_sun_distance)
                    if '6' in k_dict:
                        k1_list.append(k_dict['6'][0])
                        k2_list.append(k_dict['6'][1])
                        scale_value_list.append(k_dict['6'][2])
                        offset_value_list.append(k_dict['6'][3])
            # temperature bands Landsat 7 band 6
            elif sensor_id.lower() == cfg.sensor_etm:
                if processing_level == 'l2sp' and f[-5:-4] == '6':
                    band_names.append(files_directories.file_name(f))
                    band_number_list.append('6')
                    product_path_list.append(f)
                    sun_elevation_list.append(sun_elevation)
                    earth_sun_distance_list.append(earth_sun_distance)
                    if '6' in scale_offset_dict:
                        scale_value_list.append(scale_offset_dict['6'][0])
                        offset_value_list.append(scale_offset_dict['6'][1])
                    k1_list.append(0)
                    k2_list.append(0)
                elif processing_level == 'l1tp' and f[-12:-4] == '6_VCID_1':
                    band_names.append(files_directories.file_name(f))
                    band_number_list.append('6_VCID_1')
                    product_path_list.append(f)
                    sun_elevation_list.append(sun_elevation)
                    earth_sun_distance_list.append(earth_sun_distance)
                    if '6_VCID_1' in k_dict:
                        k1_list.append(k_dict['6_VCID_1'][0])
                        k2_list.append(k_dict['6_VCID_1'][1])
                    scale_value_list.append(1)
                    offset_value_list.append(0)
                elif processing_level == 'l1tp' and f[-12:-4] == '6_VCID_2':
                    band_names.append(files_directories.file_name(f))
                    band_number_list.append('6_VCID_2')
                    product_path_list.append(f)
                    sun_elevation_list.append(sun_elevation)
                    earth_sun_distance_list.append(earth_sun_distance)
                    if '6_VCID_2' in k_dict:
                        k1_list.append(k_dict['6_VCID_2'][0])
                        k2_list.append(k_dict['6_VCID_2'][1])
                    scale_value_list.append(1)
                    offset_value_list.append(0)
            # temperature band 10 Landsat 8-9
            elif sensor_id.lower() == cfg.sensor_oli and f[-6:-4] == '10':
                band_names.append(files_directories.file_name(f))
                band_number_list.append('10')
                product_path_list.append(f)
                sun_elevation_list.append(sun_elevation)
                earth_sun_distance_list.append(earth_sun_distance)
                if '10' in scale_offset_dict:
                    scale_value_list.append(scale_offset_dict['10'][0])
                    offset_value_list.append(scale_offset_dict['10'][1])
                k1_list.append(0)
                k2_list.append(0)
        product_name_list = [cfg.landsat] * len(band_names)
        spacecraft_list = [spacecraft_id] * len(band_names)
    if len(scale_value_list) == 0:
        scale_value_list = [0.0000275] * len(band_names)
    if len(offset_value_list) == 0:
        offset_value_list = [-0.2] * len(band_names)
    if len(k2_list) == 0:
        k2_list = [0] * len(band_names)
    if len(k1_list) == 0:
        k1_list = [0] * len(band_names)
    if len(e_sun_list) == 0:
        e_sun_list = [0] * len(band_names)
    if len(sun_elevation_list) == 0:
        sun_elevation_list = [0] * len(band_names)
    if len(earth_sun_distance_list) == 0:
        earth_sun_distance_list = [0] * len(band_names)
    processing_level_list = [processing_level] * len(band_names)
    product_date_list = [product_date] * len(band_names)
    if len(offset_value_list) == 0:
        offset_value_list = [0] * len(band_names)
    if nodata_value:
        nodata_value_list = [nodata_value] * len(band_names)
    else:
        nodata_value_list = [np.nan] * len(band_names)
    product_table = tm.add_product_to_preprocess(
        product_list=product_name_list, spacecraft_list=spacecraft_list,
        processing_level=processing_level_list,
        band_name_list=band_names, product_path_list=product_path_list,
        scale_list=scale_value_list,
        offset_list=offset_value_list, nodata_list=nodata_value_list,
        date_list=product_date_list,
        k1_list=k1_list, k2_list=k2_list, band_number_list=band_number_list,
        e_sun_list=e_sun_list, sun_elevation_list=sun_elevation_list,
        earth_sun_distance_list=earth_sun_distance_list
    )
    return product_table



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""
Band stack.

This tool allows for stacking single bands in a multiband raster.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # start the process
    >>> stack = rs.band_stack(input_bands=['path_1', 'path_2'],
    ... output_path='output_path')
"""  # noqa: E501

from typing import Union, Optional

from remotior_sensus.core import configurations as cfg
from remotior_sensus.core.bandset_catalog import BandSet
from remotior_sensus.core.bandset_catalog import BandSetCatalog
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.util import (raster_vector, shared_tools)


def band_stack(
        input_bands: Union[list, int, BandSet],
        output_path: Optional[str] = None,
        overwrite: Optional[bool] = False,
        extent_list: Optional[list] = None,
        bandset_catalog: Optional[BandSetCatalog] = None,
        n_processes: Optional[int] = None,
        virtual_output: Optional[bool] = None
):
    """Stack single bands.

    This tool allows for stacking single bands in a multiband raster.

    Args:
        input_bands: list of paths of input rasters, or number of BandSet, or BandSet object.
        output_path: string of output path.
        overwrite: if True, output overwrites existing files.
        extent_list: list of boundary coordinates left top right bottom.
        bandset_catalog: BandSetCatalog object required if input_bands is a BandSet number.
        n_processes: number of parallel processes.
        virtual_output: if True (and output_path is directory), save output
            as virtual raster.

    Returns:
        object :func:`~remotior_sensus.core.output_manager.OutputManager` with
            - path = output path

    Examples:
        Perform band stack
            >>> stack = band_stack(input_bands=['path_1', 'path_2'],
            ... output_path='output_path')
    """  # noqa: E501
    cfg.logger.log.info('start')
    cfg.progress.update(
        process=__name__.split('.')[-1].replace('_', ' '), message='starting',
        start=True
    )
    # prepare process files
    prepared = shared_tools.prepare_process_files(
        input_bands=input_bands, output_path=output_path, overwrite=overwrite,
        n_processes=n_processes, bandset_catalog=bandset_catalog,
        box_coordinate_list=extent_list, virtual_output=virtual_output
    )
    input_raster_list = prepared['input_raster_list']
    out_path = prepared['output_path']
    if input_bands is BandSet:
        bandset_x = input_bands
    elif input_bands is int:
        bandset_x = bandset_catalog.get(input_bands)
    else:
        bandset_x = BandSet.create(paths=input_raster_list)
    if virtual_output:
        virtual_path = out_path
    else:
        virtual_path = cfg.temp.temporary_file_path(name_suffix=cfg.vrt_suffix)
    raster_vector.create_virtual_raster(output=virtual_path, bandset=bandset_x)
    cfg.progress.update(message='stack', step=2, steps=2, minimum=1,
                        maximum=99, percentage=50)
    if virtual_output is not True:
        raster_vector.gdal_copy_raster(input_raster=virtual_path,
                                       output=out_path)
    cfg.progress.update(end=True)
    cfg.logger.log.info('end; band stack: %s' % str(out_path))
    return OutputManager(path=out_path)



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""Output manager.

Core class that manages several types of output, mainly intended for tools
that have several outputs.

Typical usage example:

    >>> # process output is checked
    >>> OutputManager()
"""

from remotior_sensus.core.bandset_catalog import BandSetCatalog


class OutputManager(object):
    """Manages output.

    This class manages several types of output, mainly intended for tools
    that have several outputs.
    Check argument is False if output failed.
    Single output raster or multiple file paths can be defined as arguments.
    Additional output files or tables are managed with an extra argument.
    The type of the extra argument can be flexible depending on the process 
    output.

    Attributes:
        check: True if output is as expected, False if process failed.
        path: path of the first output.
        paths: list of output paths in case of multiple outputs.
        extra: additional output elements depending on the process.

    Examples:
        Output failed
            >>> OutputManager(check=False)
            
        Output is checked and file path is provided
            >>> OutputManager(path='file.tif')
    """  # noqa: E501

    def __init__(
            self, check: bool = True, path: str = None, paths: list = None,
            extra=None
    ):
        """Initializes an Output.

        Initializes an Output. 

        Args:   
            check: True if output is as expected, False if process failed.
            path: path of the first output.
            paths: list of output paths in case of multiple outputs.
            extra: additional output elements depending on the process.

        Examples:
            Create an object with a single file path
                >>> OutputManager(path='file.tif')
                
            Create an object with several output file paths in a list and an extra argument for a dictionary
                >>> OutputManager(
                ... paths=['file1.tif', 'file2.tif'],
                ... extra={'additional_output': 'file.csv'}
                ... )
        )
        """  # noqa: E501
        self.check = check
        self.paths = paths
        if path is None:
            if paths is None:
                self.path = None
            elif len(paths) == 0:
                self.path = None
            else:
                self.path = paths[0]
        else:
            self.path = path
        self.extra = extra

    def add_to_bandset(
            self, bandset_catalog: BandSetCatalog, bandset_number=None,
            band_number=None, raster_band=None, band_name=None, date=None,
            unit=None, root_directory=None, multiplicative_factor=None,
            additive_factor=None, wavelength=None
            ):
        """Adds output to BandSet.

        Adds the OutputManager.path as a band to a BandSet in a BandSetCatalog.

        Args:
            bandset_catalog: BandSetCatalog object.
            band_name: raster name used for identifying the bands.
            wavelength: center wavelengths of band.
            unit: wavelength unit as string
            multiplicative_factor: multiplicative factor for bands during calculations.
            additive_factor: additive factors for band during calculations.
            date: date string (format YYYY-MM-DD).
            bandset_number: number of the BandSet; if None, the band is added to the current BandSet.
            root_directory: root directory for relative path.
            raster_band: raster band number.
            band_number: number of band in BandSet.

        Examples:
            Add the output to BandSet 1 as band 1.
                >>> catalog = BandSetCatalog()
                >>> OutputManager.add_to_bandset(
                ... bandset_catalog=catalog, bandset_number=1, band_number=1
                ... )
        """  # noqa: E501
        if type(bandset_catalog) is BandSetCatalog:
            bandset_catalog.add_band_to_bandset(
                path=self.path, bandset_number=bandset_number,
                band_number=band_number, raster_band=raster_band,
                band_name=band_name, date=date, unit=unit,
                root_directory=root_directory,
                multiplicative_factor=multiplicative_factor,
                additive_factor=additive_factor, wavelength=wavelength
            )
        else:
            raise Exception('bandset catalog not found')



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

"""
Tools to manage files and directories
"""

import os
import shutil
import zipfile
from pathlib import Path

from remotior_sensus.core import configurations as cfg


# check if directory exists
def is_directory(path):
    return os.path.isdir(path)


# get parent directory of input_raster
def parent_directory(path):
    try:
        directory = os.path.dirname(path)
        cfg.logger.log.debug('directory: %s' % directory)
        return directory
    except Exception as err:
        cfg.logger.log.error(str(err))
        return False


# remove directory
def remove_directory(directory):
    try:
        shutil.rmtree(directory)
    except Exception as err:
        cfg.logger.log.error(str(err))
    cfg.logger.log.debug('directory: %s' % directory)
    return directory


# remove file
def remove_file(path):
    try:
        os.remove(path)
    except Exception as err:
        cfg.logger.log.error(str(err))
    cfg.logger.log.debug('path: %s' % path)
    return path


# convert to absolute path
def relative_to_absolute_path(path, root=None):
    if root is None:
        a_path = path
    else:
        a_path = os.path.join(root, path).replace('\\', '/').replace('//', '/')
    a = Path(a_path)
    original_path = Path(path)
    if a.is_dir() or a.is_file():
        absolute = a_path
    # if absolute path is not file or directory get relative path
    elif original_path.is_dir() or original_path.is_file():
        absolute = path
    else:
        absolute = a_path
        cfg.logger.log.warning('file not found: %s' % path)
    cfg.logger.log.debug(
        'path:{}; root:{}; absolute:{}'.format(path, root, absolute)
    )
    return absolute


# create parent directory of a file path
def create_parent_directory(file_path):
    try:
        path = os.path.dirname(file_path)
        if is_directory(path):
            return path
        else:
            os.makedirs(path)
            cfg.logger.log.debug('path: %s' % path)
            return path
    except Exception as err:
        cfg.logger.log.error(str(err))
        return False


# create directory of a path
def create_directory(path):
    try:
        if not is_directory(path):
            os.makedirs(path)
            cfg.logger.log.debug('path: %s' % path)
            return path
    except Exception as err:
        cfg.logger.log.error(str(err))
        return False


# file list in directory
def files_in_directory(
        path, subdirectories=False, path_filter=None, suffix_filter=None,
        sort_files=False, root_directory=None
):
    cfg.logger.log.debug('start')
    dir_f = relative_to_absolute_path(path, root_directory)
    p = Path(dir_f)
    if subdirectories:
        pattern = '**/*'
    else:
        pattern = '*'
    if suffix_filter is not None:
        pattern = '%s%s' % (pattern, suffix_filter)
    p_list = list(p.glob(pattern))
    path_list = []
    for i in p_list:
        if i.is_file():
            if root_directory is None:
                f = i.as_posix()
            else:
                f = absolute_to_relative_path(i.as_posix(), root_directory)
            if path_filter is None:
                path_list.append(f)
            else:
                if path_filter in i.as_posix():
                    path_list.append(f)
    if sort_files:
        path_list = sorted(path_list)
    cfg.logger.log.debug('path_list: %s' % str(path_list))
    return path_list


# file extension
def file_extension(path, lower=True):
    try:
        p = Path(path)
        suffix = p.suffix
        if lower:
            suffix = suffix.lower()
        cfg.logger.log.debug('path %s; suffix: %s' % (path, suffix))
        return suffix
    except Exception as err:
        cfg.logger.log.error(str(err))
        return False


# file name
def file_name(path, suffix=False):
    try:
        p = Path(path)
        # with suffix
        if suffix:
            name = p.name
        # without suffix
        else:
            name = p.stem
        cfg.logger.log.debug('name: %s' % name)
        return name
    except Exception as err:
        cfg.logger.log.error(str(err))
        return False


# convert to relative path
def absolute_to_relative_path(path, root=None):
    p = Path(path)
    try:
        relative = p.relative_to(root)
    except Exception as err:
        cfg.logger.log.error(str(err))
        relative = p
    cfg.logger.log.debug('relative: %s' % relative.as_posix())
    return relative.as_posix()


# check if file exists
def is_file(path):
    return os.path.isfile(path)


# check output path
def output_path(path, extension):
    if not (file_extension(path) == extension):
        path = '%s%s' % (path, extension)
    return path


# check raster output path
def raster_output_path(path, virtual_output=False, overwrite=False):
    if path is None:
        path = cfg.temp.temporary_raster_path(extension=cfg.vrt_suffix)
    elif is_file(path) and not overwrite:
        raise Exception('existing path %s' % path)
    try:
        # vrt
        if virtual_output or file_extension(path) == cfg.vrt_suffix:
            virtual = True
            o_path = os.path.join(
                parent_directory(path), '{}{}'.format(
                    file_name(path, False), cfg.vrt_suffix
                ).replace('\\', '/')
            )
        # tif
        elif file_extension(path) == cfg.tif_suffix:
            virtual = False
            o_path = os.path.join(
                parent_directory(path), '{}{}'.format(
                    file_name(path, False), cfg.tif_suffix
                ).replace('\\', '/')
            )
        # other formats
        else:
            virtual = False
            o_path = os.path.join(
                parent_directory(path), '{}{}'.format(
                    file_name(path, False), cfg.tif_suffix
                ).replace('\\', '/')
            )
        cfg.logger.log.debug('o_path: %s:; virtual: %s' % (o_path, virtual))
        create_parent_directory(o_path)
        return o_path, virtual
    except Exception as err:
        cfg.logger.log.error(str(err))
        return False, False


# check input_path raster path
def input_path(path):
    if not is_file(path):
        raise Exception('file not found: ' % path)
    return path


# move file
def move_file(in_path, out_path):
    create_parent_directory(out_path)
    try:
        shutil.move(in_path, out_path)
    except Exception as err:
        cfg.logger.log.error(str(err))
    cfg.logger.log.debug('out_path: %s' % out_path)
    return out_path


# copy file
def copy_file(in_path, out_path):
    create_parent_directory(out_path)
    try:
        shutil.copy(in_path, out_path)
    except Exception as err:
        cfg.logger.log.error(str(err))
    cfg.logger.log.debug('out_path: %s' % out_path)
    return out_path


# zip file list
def zip_files(file_list, out_path, compression=None, compress_level=None):
    cfg.logger.log.debug('out_path: %s' % out_path)
    if compression is None:
        compression = zipfile.ZIP_DEFLATED
    if compress_level is None:
        compress_level = 2
    try:
        with zipfile.ZipFile(
                out_path, 'w', compression=compression,
                compresslevel=compress_level
        ) as file_zip:
            for f in file_list:
                file_zip.write(f, file_name(path=f, suffix=True))
    except Exception as err:
        cfg.logger.log.error(str(err))
    return out_path


# unzip file
def unzip_file(in_path, out_dir_path):
    file_list = []
    try:
        with zipfile.ZipFile(in_path) as file_zip:
            for name in file_zip.namelist():
                file = file_zip.open(name)
                name_f = file_name(name, suffix=True)
                output_file = open('%s/%s' % (out_dir_path, name_f), 'wb')
                file_list.append('%s/%s' % (out_dir_path, name_f))
                with file, output_file:
                    shutil.copyfileobj(file, output_file)
    except Exception as err:
        cfg.logger.log.error(str(err))
    return file_list



from unittest import TestCase

import remotior_sensus
from remotior_sensus.util import files_directories


class TestRasterToVector(TestCase):

    def test_raster_to_vector(self):
        rs = remotior_sensus.Session(
            n_processes=2, available_ram=1000, log_level=10
            )
        cfg = rs.configurations
        cfg.logger.log.debug('test')
        p = './data/S2_2020-01-01/S2_B02.tif'
        temp = cfg.temp.temporary_file_path(name_suffix=cfg.gpkg_suffix)
        cfg.logger.log.debug('>>> test raster_to_vector')
        vector = rs.raster_to_vector(p, temp)
        self.assertTrue(files_directories.is_file(vector.path))
        temp = cfg.temp.temporary_file_path(name_suffix=cfg.gpkg_suffix)
        cfg.logger.log.debug('>>> test raster_to_vector dissolve')
        vector = rs.raster_to_vector(p, temp, dissolve=True)
        self.assertTrue(files_directories.is_file(vector.path))

        # clear temporary directory
        rs.close()
